require 'fake-intrinsics'

raygen = require('saotrace.raygen' )

vobj = require('va').vobj

vspec = require( 'saotrace.raygen.jitter' ).vspec

describe( "lissa-jous", function ()

   it( 'missing args', function()

       local ok, args = vobj:validate( vspec.lissajous,
				      { name = 'lj0' }
				   )
       assert.is_false( ok )
       assert.matches( ': required but not specified', args )
    end)


   it( 'incorrect pitch', function()
       local ok, args = vobj:validate( vspec.lissajous,
				  { name = 'lj0',  pitch = 3 }
				  )
       assert.is_false( ok )
       assert.matches( 'pitch: incorrect type', args )
    end)

   it( 'missing yaw', function()
       local ok, args = vobj:validate( vspec.lissajous,
				  { name = 'lj0',
				    pitch = { period = 3, amp = 2} }
				  )
       assert.is_false( ok )
       assert.matches( 'yaw: required', args )
    end)

   it( 'incorrect yaw', function()

       local ok, args = vobj:validate( vspec.lissajous,
				  { name = 'lj0',
				    pitch = { period = 3, amp = 2}, yaw = 3 }
				  )
       assert.is_false( ok )
       assert.matches( 'yaw: incorrect type', args )

    end)

    it( 'correct (1)', function()
       local ok, args = vobj:validate( vspec.lissajous,
				       { name = 'lj0',
					 pitch = { period = 3, amp = { 3, 'radian' }, phase = 3},
					 yaw   = { period = 3, amp = { 3, 'radian' }, phase = 3}
				      }
				  )
       assert.is_true( ok )

       assert.equals( 'lj0', args.name )
       assert.equals( 3, args.pitch.phase[1] )
       assert.equals( 3, args.pitch.amp[1] )
       assert.equals( 3, args.pitch.period[1] )
       assert.equals( 3, args.yaw.phase[1] )
       assert.equals( 3, args.yaw.amp[1] )
       assert.equals( 3, args.yaw.period[1] )

    end)

    it( 'correct (2)', function()

       local ok, args = vobj:validate( vspec.lissajous,
				       { name = 'lj0',
					 pitch = { phase = 3, amp = { 4, 'radian' }, period = 5},
					 yaw   = { phase = 6, amp = { 7, 'radian' }, period = 8 }
				      }
				  )
       assert.is_true( ok )

       assert.equals( 3, args.pitch.phase[1] )
       assert.equals( 4, args.pitch.amp[1] )
       assert.equals( 5, args.pitch.period[1] )
       assert.equals( 6, args.yaw.phase[1] )
       assert.equals( 7, args.yaw.amp[1] )
       assert.equals( 8, args.yaw.period[1] )

    end)

end)

describe( "table", function ()

    it( 'quat_col, optical_axis, input_origin', function()


	local ok, args = vobj:validate( vspec.file,
					{
					   name = 'f0',
					   file = 'foo.fits',
					   quat_col = 'q_attr',
					   optical_axis = {
					      { angle = 1, axis = { 0, 0, 1 } },
					      { angle = 2, axis = { 0, 1, 0 } },
					      { angle = 3, axis = { 1, 1, 1 } },
					   },

					   input_origin = {
					      { angle = 3, axis = { 1, 1, 1 } },
					      { angle = 2, axis = { 0, 1, 0 } },
					      { angle = 1, axis = { 1, 0, 0 } },
					   },
					}
				     )

	assert.is_true( ok )

	local axis = args.optical_axis
	assert.equals( 1, axis[1].angle[1], "optical_axis[1].angle" )
	assert.equals( 2, axis[2].angle[1], "optical_axis[2].angle" )
	assert.equals( 3, axis[3].angle[1], "optical_axis[3].angle" )
	assert.are.same( { 0, 0, 1}, axis[1].axis, "optical_axis[1].axis" )
	assert.are.same( { 0, 1, 0}, axis[2].axis, "optical_axis[2].axis" )
	assert.are.same( { 1, 1, 1}, axis[3].axis, "optical_axis[3].axis" )

	local axis = args.input_origin
	assert.equals( 3, axis[1].angle[1], "input_origin[1].angle" )
	assert.equals( 2, axis[2].angle[1], "input_origin[2].angle" )
	assert.equals( 1, axis[3].angle[1], "input_origin[3].angle" )
	assert.are.same( { 1, 1, 1}, axis[1].axis, "input_origin[1].axis" )
	assert.are.same( { 0, 1, 0}, axis[2].axis, "input_origin[2].axis" )
	assert.are.same( { 1, 0, 0}, axis[3].axis, "input_origin[3].axis" )

     end)


    it( "axes( ra, dec, roll)", function()

	local ok, args = vobj:validate( vspec.file,
				      {
					 name = 'f0',
					 file = 'foo.fits',
					 axes = {
					    { col = 'ra',   axis = { 0,  0, 1 } },
					    { col = 'dec',  axis = { 0, -1, 0 } },
					    { col = 'roll', axis = { 1,  0, 0 } },
					 }

				      }
				     )


	assert.is_true( ok, args, args )

	local axis = args.axes
	assert.equals( 'ra',   axis[1].col, "axis[1].col" )
	assert.equals( 'dec',  axis[2].col, "axis[2].col" )
	assert.equals( 'roll', axis[3].col, "axis[3].col" )

	assert.are.same( { 0,  0, 1}, axis[1].axis, "axis[1].axis" )
	assert.are.same( { 0, -1, 0}, axis[2].axis, "axis[2].axis" )
	assert.are.same( { 1,  0, 0}, axis[3].axis, "axis[3].axis" )


     end)


    it( "axes( ra, dec)", function()

	local ok, args = vobj:validate( vspec.file,
				      {
					 name = 'f0',
					 file = 'foo.fits',
					 axes = {
					    { col = 'ra',   axis = { 0,  0, 1 } },
					    { col = 'dec',  axis = { 0, -1, 0 } },
					 }

				      }
				     )

	assert.is_false( ok )
	assert.matches( "expected 3.+got 2", args )
     end)

 end)
