// -*-C++-*-
#ifndef JITTER_COMMON_H
#define JITTER_COMMON_H

#include <string>

#include <Eigen/Core>
#include <Eigen/Geometry>

namespace Jitter {

    struct Point {
	double time;
	Eigen::Quaterniond q;
    };

    struct EulerAxis {

	std::string name;
	std::string colname;
	Eigen::Vector3d axis;
	double scale;
	bool autoscale;

	EulerAxis() {}

	EulerAxis( const char* name,
		   const char* colname,
		   Eigen::Vector3d& axis,
		   double s ) :
	    name( name ), colname(colname),
	    axis(axis), scale(s), autoscale(true) {}

	EulerAxis( const char* name,
		   const char* colname,
		   Eigen::Vector3d axis,
		   double s ) :
	    name( name ), colname(colname),
	    axis(axis), scale(s), autoscale(true) {}

    };

    //---------------------------------------------------------------------
    // Function object to apply a transformation to a rotation

    struct ApplyStaticXfrm {
	Eigen::Matrix3d xfrm;

	ApplyStaticXfrm (Eigen::Matrix3d xfrm ) : xfrm(xfrm) {}

	void operator() ( Point& p ) const {

	    // the transformation is applied directly to the vector
	    // part of the quaternion.  It had better not change the
	    // vector's normalization!
	    p.q.vec() = xfrm * p.q.vec();
	}
    };


    //---------------------------------------------------------------------
    // Function object to apply a constant post-rotation

    struct ApplyPostQuat {

	Eigen::Quaterniond post;
	ApplyPostQuat ( Eigen::Quaterniond q ) : post( q ) {}

	void operator() ( Point& p )  const {

	    p.q = post * p.q;
	}
    };

    //---------------------------------------------------------------------
    // Function object to transform a sequence of additive quaternions
    // to quaternions relative to the start of the sequence

    struct AdditiveToAbsolute {

	Eigen::Quaterniond prev;
	AdditiveToAbsolute ( Eigen::Quaterniond q ) : prev( q ) {}

	void operator() ( Point& p )  {

	    Eigen::Quaterniond q = p.q;

	    p.q = prev * p.q;

	    prev = q;
	}
    };

    //---------------------------------------------------------------------
    // Function object to create a sequence of additive rotations from
    // a sequence of absolute rotations

    struct AbsoluteToAdditive {

	Eigen::Quaterniond prev;
	AbsoluteToAdditive (Point& prev ) : prev( prev.q ) {}

	void operator() ( Point& p )  {

	    Eigen::Quaterniond q = p.q;

	    p.q = p.q * prev.inverse();

	    prev = q;
	}
    };


}

#endif // !JITTER_COMMON_H
