#include <string>
#include <sstream>

#include <fitsio.h>

#include <tracefctxx/TraceFct.h>
#include <Exception/Exception.h>

#include "units.h"
#include "utils.h"

#include <Eigen/SVD>
using namespace Eigen;

#include "Jitter/File.h"

#include "SimpleFits.h"

using namespace raygen;

namespace Jitter {

    namespace TableFile {

	//---------------------------------------------------------------------
	size_t
	Base::read( Point*& points, TimeInterval& tvalid ) {

	    TraceFct tf( "Jitter::TableFile::read" );

	    size_t nrows;

	    if ( format == "auto" && ( raygen::file_has_ext( file, ".fits" ) || raygen::file_has_ext( file, ".fits.gz" ) ) )
		format = "fits";

	    if ( format == "fits" ) {

		tf.message( "jitter: reading data from %s...\n", file.c_str() );

		nrows = read_fits( points, tvalid );

		tf.message( "jitter: read %d rows\n", nrows );

	    }

	    else
	    {
		throw( Exception( "jitter input file( " + file + "): can't figure out file format" ) );
	    }

	    return nrows;
	}

	//---------------------------------------------------------------------
	size_t
	Quaternion::read_fits( Point*& points, TimeInterval& tvalid ) {

	    SimpleFits::Table table( file );

	    rowsize = rowsize && rowsize < table.rowsize() ? rowsize : table.rowsize();
	    points = new Point[table.num_rows()];

	    // get info on required columns
	    SimpleFits::Table::Column time_col = table.column( "time", time_col_name );
	    SimpleFits::Table::Column quat_col = table.column( "quaternion", quat_col_name );

	    if ( quat_col.repeat != 4 ) {

		ostringstream ost;

		ost << "expected 4 elements per cell in " << quat_col_name << ", got " << quat_col.repeat;

		throw( Exception( ost.str() ) );
	    }


	    // these are allocated by table.read_col_double the first time through
	    double* quat = NULL;
	    double* time = NULL;

	    long row = 1;
	    LONGLONG nread;
	    for( row = 1;  ( nread = table.read_col_double( time, time_col, row, rowsize ) ) ; row += nread ) {

		LONGLONG nread2 = table.read_col_double( quat, quat_col, row, rowsize );

		tf_assert( nread == nread2 );

		Point*  point = &points[row-1];
		double* quat_t = quat;
		for ( LONGLONG irow = 0 ; irow < nread ; irow++, quat_t += quat_col.repeat ) {

		    point[irow].time = time[irow];
		    //		point[irow].q = Quaterniond( quat_t[3], quat_t[0], quat_t[1], quat_t[2] );
		    point[irow].q = Quaterniond( quat_t );
		}

	    }

	    if ( ! table.read_keyword( "TSTART", tvalid.min ) )
	      tvalid.min = points[0].time;

	    if ( ! table.read_keyword( "TSTOP", tvalid.max ) )
	      tvalid.max = points[table.num_rows()-1].time;

	    delete [] quat;
	    delete [] time;

	    return table.num_rows();

	}

	//---------------------------------------------------------------------

	size_t
	Euler::read_fits( Point*& points, TimeInterval& tvalid ) {

	    SimpleFits::Table table( file );

	    rowsize = rowsize && rowsize < table.rowsize() ? rowsize : table.rowsize();

	    points = new Point[table.num_rows()];

	    // get info on required columns
	    SimpleFits::Table::Column time_col  = table.column( "time", time_col_name );
	    SimpleFits::Table::Column rot[3];

	    // these are allocated by table.read_double the first time through
	    double* rotd[3] = { NULL, NULL, NULL };
	    double* time  = NULL;

	    for ( size_t irot = 0 ; irot < 3 ; irot++ ) {

		EulerAxis* axis = axes[irot];
		rot[irot] = table.column( axis->name, axis->colname );

		axis->scale = 1.0;
		if ( axis->autoscale && ! raygen::Units::convert_angular( rot[irot].unit, axis->scale ) )
		    throw( Exception( axis->name + " column \"" + axis->colname + "\": unknown units: " + rot[irot].unit ) );
	    }

	    long row = 1;
	    LONGLONG nread;

	    for( row = 1;  ( nread = table.read_col_double( time, time_col, row, rowsize ) ) ; row += nread ) {

		LONGLONG nread2;

		for( size_t irot = 0 ; irot < 3 ; irot++ ) {

		    nread2 = table.read_col_double( rotd[irot], rot[irot], row, rowsize );
		    tf_assert( nread == nread2 );

		}

		Point*  point = &points[row-1];
		for ( LONGLONG irow = 0; irow < nread ; irow++ ) {

		    point[irow].time = time[irow];


#ifndef RA_and_DEC
		    point[irow].q =
			  Quaterniond( AngleAxisd( axes[0]->scale * rotd[0][irow], axes[0]->axis ) )
			* Quaterniond( AngleAxisd( axes[1]->scale * rotd[1][irow], axes[1]->axis ) )
			* Quaterniond( AngleAxisd( axes[2]->scale * rotd[2][irow], axes[2]->axis ) )
			;

#else // hardwire RA & Dec

		    double yaw   = axis[0]->scale * rotd[0][irow];
		    double pitch = axis[1]->scale * rotd[1][irow];
		    double roll  = axis[2]->scale * rotd[2][irow];

		    Quaterniond   Qyaw( cos(yaw/2),    0,           0,             sin(yaw/2) );
		    Quaterniond Qpitch( cos(-pitch/2), 0,           sin(-pitch/2), 0 );
		    Quaterniond  Qroll( cos(roll/2),   sin(roll/2), 0,             0 );

		    point[irow].q = Qyaw * Qpitch * Qroll;
#endif

		}

	    }

	    if ( ! table.read_keyword( "TSTART", tvalid.min ) )
	      tvalid.min = points[0].time;

	    if ( ! table.read_keyword( "TSTOP", tvalid.max ) )
	      tvalid.max = points[table.num_rows()-1].time;

	    for( size_t irot = 0; irot < 3 ; irot++ )
		delete [] rotd[irot];

	    delete [] time;

	    return table.num_rows();
	}

    }

}
