#include <lua.hpp>

#include <tracefctxx/TraceFct.h>
#include <suplib/str.h>

#include "LuaState.h"

#include "lua_raygen.h"
#include "lua_system.h"

typedef struct lua_State luaState;

// this is really C code; until we catch exceptions properly in the
// calling code, can't avoid it.

static char *
stack_top_msg( lua_State *L )
{
  const char *topmsg;
  char *msg;

  tf_assert( lua_gettop( L ) && (topmsg = lua_tostring(L, -1)) );
  tf_assert( msg = str_dup( topmsg ) );
  return msg;
}

static
int panic( lua_State *L ) {

    TraceFct::die( "%s", stack_top_msg( L ) );

    /* NOT REACHED (hopefully) */
    return 1;
}

static int
stash_package_loaded( luaState *L ) {
    TraceFct tf( __func__ );

    lua_getglobal( L, "package" );	// l = 1

    if ( lua_isnil( L, -1 ) ) {
	lua_pop( L, 1 );		// l = 0
	tf.die( "internal error: unable to find Lua package table" );
    }

    lua_getfield( L, -1, "loaded" );	// l = 2
    if ( lua_isnil( L, -1 ) ) {
	lua_pop( L, 2 );		// l = 0
	tf.die( "internal error: unable to find Lua package.loaded table" );
    }

    lua_remove( L, -2 );		// l = 1

    return luaL_ref( L, LUA_REGISTRYINDEX );
}

void
LuaState::get_package_loaded( ) {
    lua_rawgeti( L, LUA_REGISTRYINDEX, package_loaded );
}


LuaState::LuaState() {

    TraceFct tf( "LuaState::LuaState" );

    /* create interpreter instance */
    if ( NULL == ( L = luaL_newstate() ) )
        tf.die( "error creating Lua interpreter" );

    /* register a new panic function */
    lua_atpanic( L, panic );

    /* register standard lua libraries */
    luaL_openlibs( L );

    // grab reference to package.loaded
    package_loaded = stash_package_loaded( L );

    /* register raygen modules, both compiled and pure Lua */
    luaopen_raygen( *this );
    luaopen_raygen_system( *this );

    // load the configuration system
    require_module( "saotrace.config" );

}

LuaState::~LuaState() {

    luaL_unref( L, LUA_REGISTRYINDEX, package_loaded );
    lua_close( L );

}

void
LuaState::get_module_attr( const char* module, const char* attr ) {

    require_module( module, true );

    lua_getfield( L, -1, attr );	// l = 2

    lua_remove( L, -2 );		// l = 1
}

// set module.attr = top of stack
void
LuaState::set_module_attr( const char* module, const char* attr ) {

    require_module( module, true );	// l = 2 ; include attr value on stack

    // swap attr value and module table
    lua_insert( L, -2 );
    // stack is now 1: module table, 2: attr value

    lua_setfield( L, -2, attr );	// l =  1; pops attribute value

    lua_pop( L, 1 );			// l =  0; pop module table
}

// register a null terminated list luaL_Reg structs
void
LuaState::register_module ( const char* module, luaL_Reg* l ) {

    get_package_loaded( );		// l = 1

#if   LUA_VERSION_NUM == 501

    lua_newtable( L );			// l = 2
    luaL_register( L, NULL, l );

#elif LUA_VERSION_NUM > 501

    luaL_newlib( L, l );		// l = 2

#else

# error "unsupported Lua version"

#endif

    lua_setfield( L, -2, module );	// l = 1
    lua_pop( L, 1 );			// l = 0

}

// register table on stack
void
LuaState::register_module( const char* module  ) {

    get_package_loaded( );		// l = 1
    lua_insert( L, -2 );		// swap top two elements

    // stack is now 1: package.loaded, 2: table

    lua_setfield( L, -2, module );	// l = 1
    lua_pop( L, 1 );			// l = 0
}

void
LuaState::register_module( const char* module, luaL_Reg* fptr[] )
{
    size_t nfunc = 0;
    luaL_Reg** ptr = fptr;
    luaL_Reg* func;
    luaL_Reg* funcp;

    /* count functions */

    while( *ptr) ptr++;

    nfunc = ptr - fptr;
    func = funcp = (luaL_Reg*) malloc( ( nfunc + 1 ) * sizeof(luaL_Reg) );

    while ( nfunc-- )
	*funcp++ = **fptr++;

    (*funcp).name = NULL;
    (*funcp).func = NULL;

    register_module( module, func );
    free( func );
}


void
LuaState::require_module( const char* module, bool push_module ) {

    lua_getglobal( L, "require" );
    lua_pushstring( L, module );
    if ( lua_pcall( L, 1, push_module ? 1 : 0, 0) ) {
	TraceFct tf( __func__ );
	tf.die( "error loading %s: %s", module, lua_tostring(L, -1 ) );
    }

}


void
LuaState::call_module_func( const char* module,
			    const char* func,
			    int nargs,
			    int nresults
			    ) {

    get_module_attr( module, func );
    if ( lua_isnil( L, -1 )  ) {
	TraceFct tf( __func__ );
	tf.die( "unable to find module function %s.%s", module, func );
    }

    lua_insert( L, -(nargs+1) );
    lua_call( L, nargs, nresults );

}
