#include <string>
#include <sstream>

#include "SimpleFits.h"


namespace SimpleFits {

    Exception
    File::exception( string pfx ){

	char errmsg[81];
	fits_read_errmsg( errmsg );
	return Exception( (pfx.size() ? pfx + ": " : "" ) + errmsg + " in " + file );
    }

    File::File ( const string& file ) : fptr( NULL ), status( 0 ), file( file ) {

	if ( fits_open_file( &fptr, file.c_str(), READONLY, &status ) )
	    throw( exception() );
    }

    File::~File( void ) {

	if ( fptr ) {

	    // ignore previous status.  we need to get out of here
	    int status = 0;
	    fits_close_file( fptr, &status );

	    if ( status )
		throw( exception( "error closing fits file: " + file ) );
	}
    }


    string
    File::read_keyword( const char* keyname ) throw( Exception ) {

	char value[80];

	if ( fits_read_key( fptr, TSTRING, keyname, &value, NULL, &status ) )
	    throw exception( string(keyname) + ": ");

	return string(value);
    }

    bool
    File::read_keyword( const char* keyname, double& value ) {

	// don't change object status value as the keyword need not
	// exist
	int s_status = status;

	fits_read_key( fptr, TDOUBLE, keyname, &value, NULL, &s_status );

	return s_status == 0;
    }

    double
    File::read_keyword_double( const char* keyname ) throw( Exception ) {

	double value;

	if ( fits_read_key( fptr, TDOUBLE, keyname, &value, NULL, &status ) )
	    throw exception( string(keyname) + ": ");

	return value;
    }

    Table::Table( const string& file ) {

	this->file = file;

	if ( fits_open_table( &fptr, file.c_str(), READONLY, &status ) )
	    throw( exception() );

	if ( fits_get_num_rows( fptr, &_nrows, &status ) )
	    throw( exception() );

	if ( fits_get_rowsize( fptr, &_rowsize, &status ) )
	    throw( exception() );
    }

    bool
    Table::exists_column( const string& column_name ) {

	int status = 0;
	int colnum;

	fits_get_colnum( fptr, 0, const_cast<char *>(column_name.c_str()), &colnum, &status );
	return status == 0;
    }

    LONGLONG
    Table::read_col_double( double* &dst, const Column& col, LONGLONG start_row, LONGLONG nrows ) {

	LONGLONG nread = start_row + nrows -1 > num_rows() ? num_rows() - start_row + 1
                                                           : nrows;

	if ( dst == NULL )
	    dst = new double[ nread * col.repeat ];

	if ( fits_read_col( fptr, TDOUBLE, col.colnum, start_row,
			    1, nread * col.repeat,
			    NULL, dst,
			    NULL,
			    &status )
	     ) {

	    throw( exception( file + ": error reading elements" ) );
	}

	return nread;
    }


    Table::Column::Column( Table&  tbl, const string& alias, const string& colname ) {

	int status = 0;

	if ( fits_get_colnum( tbl.fptr, 0, const_cast<char *>(colname.c_str()), &colnum, &status ) )
	    throw( tbl.exception( alias + " column \"" + colname + '"' ) );

	// fits_get_coltype wants long (not LONGLONG) for repeat
	long lrepeat;

	if ( fits_get_coltype( tbl.fptr, colnum, &typecode, &lrepeat, &width, &status ) )
	    throw( tbl.exception( alias + " column \"" + colname + '"' ) );

	repeat = lrepeat;

	ostringstream ost;
	char value[80];
	ost << "TUNIT" << colnum ;

	if ( fits_read_key_str( tbl.fptr, ost.str().c_str(), value, NULL, &status )
	     && status != KEY_NO_EXIST
	     )
	    throw( tbl.exception( alias + " column \"" + colname + '"' ) );

	unit = value;

	name = colname;

    }

    ostream& operator<<( std::ostream& ostr, const Table::Column& c ) { 

	ostr << "typecode: " << c.typecode << "\n"
	     << "colnum:   " << c.colnum << "\n"
	     << "name:     " << c.name << "\n"
	     << "repeat:   " << c.repeat << "\n"
	     << "unit:     " << c.unit << "\n"
	     << "width:    " << c.width << "\n";

	return ostr;

    }


}
