/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *       Free Software Foundation, Inc.
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <math.h>
#include <string.h>

#include <randlib_c/randlib.h>

#include <tracefctxx/TraceFct.h>
#include <tracefct/exiterrvals.h>

#include <mathconst/mathconst.h>

#include <lua.hpp>
#include <lauxlib.h>

#include "main.h"
#include "lua_stack.h"
#include "misc.h"
#include "units.h"

#include "sourcepos.h"

using Eigen::Quaterniond;
using Eigen::AngleAxisd;
using Eigen::Vector3d;

#include "pos_gen.h"
#include "dist.h"

typedef struct
{
  double	radius;		/* radius of disk */

  ImgDist	idist;

  SourcePosType pos_type;
  SourcePos     pos;          /* position of source */
  double	z;	      /* z coord of disk */

} DiskData;

static int	define(lua_State *);
static int	init_angular(PositionGeneratorData *data);
static int	generate_angular( void *data, DVector3 *direction, double *z );
static void	cleanup(void *data);

luaL_Reg extent_disk = { "disk", define };

static PositionGeneratorSpec spec_angular =
{
  "disk",
  init_angular,
  generate_angular,
  cleanup
};

static int
define( lua_State* L )
{
  DiskData *data;
  const char* name;
  char err_pfx[1024];

  TraceFct tf( "disk" );

  name = get_lua_cstring( L, "name" );

  /* generate error message prefix */
  sprintf( err_pfx, "disk `%s'", name );

  if ( NULL == ( data = (DiskData *) malloc( sizeof(DiskData) ) ) )
      luaL_error( L, "%s: error allocating memory", err_pfx );

  data->z = get_src_pos( L, err_pfx, &(data->pos_type), &(data->pos) );

  if ( data->pos_type != SOURCE_POS_ANGULAR )
      luaL_error( L, "%s: unsupported source position type", err_pfx );

  /* now that we know the axial location, process the radius,
     converting into angular units if necessary */
  data->radius = get_lua_extent( L, err_pfx, "radius", 
				 data->z, raygen::Units::sys().radian() );
  if ( data->radius <= 0.0 )
      luaL_error( L, "%s: radius must be positive", err_pfx );

  /* determine  intensity distribution */
  get_ImgDist( L, err_pfx, &data->idist, data->z,
	       raygen::Units::sys().radian(),
	       -data->radius, data->radius,
	       -data->radius, data->radius );

  new_pos_gen( &spec_angular, name, data );

  return 0;
}


static int
init_angular( PositionGeneratorData *data )
{
  DiskData *disk = (DiskData *) data->data;

  data->z		= disk->z;
  data->factor		= 1.0;
  data->max_extent	= disk->radius + disk->pos.angular.theta;
  return 0;
}

static int
generate_angular( void *data, DVector3 *direction, double *z )
{
  DiskData *disk = (DiskData *) data;
  SourcePosAngular *pos = &(disk->pos.angular);
  double r2, dx, dy;
  *z		= pos->z;

  r2 = disk->radius * disk->radius;
  do
  {
      gen_ImgDist( &disk->idist, &dx, &dy );
  }
  while( dx * dx + dy * dy > r2 );

  // dx and dy are in the direction of the source.  since the direction
  // vector is from the source to the telescope, must flip them.

  // positive rotation about the X axis takes Y into Z, so a positive
  // dy makes the direction vector point in a negative Y direction,
  // which is what we want.

  // positive rotation about the Y axis takes Z into X, so a positive
  // dx makes the direction vector point in a *positive* X direction,
  // which is opposite of what we want.

  Vector3d v =
    *pos->pos
    * AngleAxisd( dy, Vector3d::UnitX())
    * AngleAxisd( -dx, Vector3d::UnitY())
    * Vector3d::UnitZ();

  direction->x = v.x();
  direction->y = v.y();
  direction->z = v.z();

  return 0;
}

static void
cleanup( void *data )
{
  free ( data );
}
