/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *       Free Software Foundation, Inc.
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <iostream>

#include <stdlib.h>

#include <randlib_c/randlib.h>
#include <lua.hpp>
#include <lauxlib.h>

#include <suplib/str.h>
#include <suplib/debug.h>

#include "dist.h"
#include "lua_stack.h"
#include "misc.h"

static TokListToken
dist[] =
{
    { "gaussian", DT_GAUSSIAN },
    { "uniform",  DT_UNIFORM }
};
static TokList dists = GenTokList( dist );


/*******************************************************************/
/*******************************************************************/

/* These generate the random deviates */

static void
uniform1D( void *data, double *x )
{
    Dist1D *dist = (Dist1D *) data;
    *x = genunf( dist->min, dist->max );
}

static void
gaussian1D( void *data, double *x )
{
    Dist1D *dist = (Dist1D *) data;
    Gauss1D *g1d = &(dist->dist.gauss);
    double val;

    do
      val = gennor( g1d->mean, g1d->sigma );
    while ( val < dist->min || val > dist->max );

    *x = val;
}

static void
uniform2D( void *data, double *x, double *y )
{
    Dist2D *dist = (Dist2D *) data;
    *x = genunf( dist->xmin, dist->xmax );
    *y = genunf( dist->ymin, dist->ymax );
}

static void
gaussian2D( void *data, double *x, double *y )
{
    Dist2D *dist = (Dist2D *) data;
    Gauss2D *g2d = &(dist->dist.gauss);

    double p[2];
    double work[2];

    do
	genmn( g2d->gparm, p, work );
    while ( p[0] < dist->xmin || p[0] > dist->xmax ||
	    p[1] < dist->ymin || p[1] > dist->xmax );

    *x = p[0];
    *y = p[1];
}


/*******************************************************************/
/*******************************************************************/

/* extract the deviate parameters from the Lua stack */

/*******************************************************************/
static void
extract_gaussian1D( lua_State *L, const char* err_pfx, Gauss1D *g1d, double z, ut_unit* unit )
{
    g1d->mean  = get_lua_extent( L, err_pfx, "mean", z, unit );
    g1d->sigma = get_lua_extent( L, err_pfx, "sigma", z, unit );
}

static void
extract1D( lua_State *L, const char *err_pfx, Dist1D *dist, double z, ut_unit* unit )
{
    const char *str = lua_tostring( L, -1 );
    int type = tokmatch( str, &dists );

    switch( type )
    {
    case DT_GAUSSIAN:
	dist->gen = gaussian1D;
	extract_gaussian1D( L, err_pfx, &(dist->dist.gauss), z, unit );
	break;

    case DT_UNIFORM:
	dist->gen = uniform1D;
	break;

    default:
	/* shouldn't get here! */
	luaL_error( L, "%s: internal error, unknown distribution: `%s'",
		    err_pfx, str_dup(str) );
    }
    lua_pop(L, 1 );
}

/*******************************************************************/

static void
extract_gaussian2D( lua_State *L, const char* err_pfx, Gauss2D *g2d, double z, ut_unit* unit )
{
    char *str;

    g2d->meanv[0] = get_lua_extent( L, err_pfx, "x", z, unit );
    g2d->meanv[1] = get_lua_extent( L, err_pfx, "y", z, unit );


    /* single sigma, or sigma_x, sigma_y? */
    str = get_lua_cstring( L, "sigma" );
    if ( str )
    {
	g2d->covm[0] = g2d->covm[3] = get_lua_extent( L, err_pfx, "sigma", z, unit );
	free( str );
    }
    else
    {
	str = get_lua_cstring( L, "sigma_x" );
	g2d->covm[0] = get_lua_extent( L, err_pfx, "sigma", z, unit );
	free( str );

	str = get_lua_cstring( L, "sigma_y" );
	g2d->covm[3] = get_lua_extent( L, err_pfx, "sigma", z, unit );
	free( str );
    }

    g2d->covm[0] *= g2d->covm[0];
    g2d->covm[1] = 0;
    g2d->covm[2] = 0;
    g2d->covm[3] *= g2d->covm[3];

    setgmn( g2d->meanv, g2d->covm, 2, g2d->gparm );
}

static void
extract2D( lua_State *L, const char *err_pfx, Dist2D *dist, double z, ut_unit* unit )
{
    const char *str = lua_tostring( L, -1 );
    int type = tokmatch( str, &dists );

    dist->type = (DistType) type;

    switch( type )
    {
    case DT_GAUSSIAN:
	dist->gen = gaussian2D;
	extract_gaussian2D( L, err_pfx, &(dist->dist.gauss), z, unit );
	break;

    case DT_UNIFORM:
	dist->gen = uniform2D;
	break;

    default:
	/* shouldn't get here! */
	luaL_error( L, "%s: internal error, unknown distribution: `%s'",
		    err_pfx, str_dup(str) );
    }
    lua_pop(L, 1 );
}

/******************************************************************************/

static void
dump_ImgDist( ImgDist* idist ) {

    if ( idist->twoD ) {

	std::cerr << "twoD" << std::endl;

	Dist2D& d = idist->dist.xy;

	std::cerr << "x min: " << d.xmin << std::endl;
	std::cerr << "x max: " << d.xmax << std::endl;
	std::cerr << "y min: " << d.ymin << std::endl;
	std::cerr << "y max: " << d.ymax << std::endl;


    }
    else {

	std::cerr << "oneD" << std::endl;

	Dist1D& d = idist->dist.x_y.x;

	std::cerr << "x min: " << d.min << std::endl;
	std::cerr << "x max: " << d.max << std::endl;

	d = idist->dist.x_y.y;

	std::cerr << "y min: " << d.min << std::endl;
	std::cerr << "y max: " << d.max << std::endl;


    }
}


void
get_ImgDist( lua_State *L, const char *err_pfx,
	     ImgDist *idist, double z, ut_unit* unit,
	     double xmin, double xmax, double ymin, double ymax )
{
  /* check if an intensity distribuion was specified */
  lua_getfield( L, -1, "idist" );

  if ( ! lua_isnil( L, -1 ) )
  {
      /* if it's a 2D distribution, there will be a list element at [1]*/
      lua_pushinteger( L, 1 );
      lua_gettable( L, -2 );
      idist->twoD = ! lua_isnil( L, -1 );

      if ( idist->twoD )
      {
	  extract2D( L, err_pfx, &(idist->dist.xy), z, unit );
	  idist->dist.xy.xmin = xmin;
	  idist->dist.xy.xmax = xmax;
	  idist->dist.xy.ymin = ymin;
	  idist->dist.xy.ymax = ymax;
      }
      else
      {
	  lua_pop( L, 1 );

	  lua_getfield( L, 1, "x" );
	  extract1D( L, err_pfx, &(idist->dist.x_y.x), z, unit );
	  lua_pop(L, 1 );

	  idist->dist.x_y.x.min = xmin;
	  idist->dist.x_y.x.max = xmax;

	  lua_getfield( L, 1, "y" );
	  extract1D( L, err_pfx, &(idist->dist.x_y.y), z, unit );
	  lua_pop(L, 1 );

	  idist->dist.x_y.y.min = xmin;
	  idist->dist.x_y.y.max = xmax;
      }

  } else {

      idist->twoD = 1;
      idist->dist.xy.type = DT_UNIFORM;
      idist->dist.xy.gen = uniform2D;

      idist->dist.xy.xmin = xmin;
      idist->dist.xy.xmax = xmax;
      idist->dist.xy.ymin = ymin;
      idist->dist.xy.ymax = ymax;
  }

  lua_pop( L, 1 );
}

void
gen_ImgDist( ImgDist *idist, double *x, double *y )
{
    if ( idist->twoD )
	idist->dist.xy.gen( &(idist->dist.xy), x, y );
    else
    {
	idist->dist.x_y.x.gen( &(idist->dist.x_y.x), x );
	idist->dist.x_y.y.gen( &(idist->dist.x_y.y), y );
    }
}

