/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2010 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * -->8-->8-->8-->8-- */

#include <math.h>

#include <lua.hpp>
#include <lauxlib.h>

#include <mathconst/mathconst.h>
#include <tracefct/exiterrvals.h>
#include <tracefctxx/TraceFct.h>
#include <randlib_c/randlib.h>
#include <suplib/str.h>

#include "main.h"

#include "ea.h"

#include "common.h"
#include "lua_stack.h"

/*
  the list of entrance apertures created by the lua routines.  This
  is eventually replaced by the "ea" array in the System structure.
*/
static SLinkList ea_list;


/*-------------------------------------*/
/* central entrance aperture creation */

void
new_entrance_aperture( lua_State *L,
		       const char* errpfx,
		       void *data,
		       unsigned int id,
		       double area,
		       EA_Pos  pos,
		       EA_Cleanup cleanup )
{
  EntranceAperture *ea;

  if ( NULL == ( ea = (EntranceAperture *) malloc( sizeof(EntranceAperture) ) ) )
      luaL_error( L, "%s: error allocating memory for entrance aperture", errpfx );

  ea->data    = data;
  ea->pos     = pos;
  ea->cleanup = cleanup;
  ea->area    = area;
  ea->id      = id;

  if ( sll_insert_head( ea_list, ea ) )
      luaL_error( L, "%s: error installing entrance aperture", errpfx );

  sys.ea_area += area;
}

/*-------------------------------------*/


void
ea_setup( void)
{
    TraceFct tf( "ea_init" );
    if ( NULL == ( ea_list = sll_new( SLL_NULL_CMP ) ) )
	tf.exit( ExitERR_alloc, "unable to initialize entrance ap list" );

    sys.ea_area = 0;
}



void
ea_init( void ) {
    TraceFct tf( "ea_process" );

  sys.n_ea = sll_count( ea_list );
  /* make sure that there's at least one entrance aperture out there */
  if ( 0 == sys.n_ea )
      tf.die( "no entrance apertures were specified" );

  /* now, create an array of entrance apertures */
  if ( NULL ==
       ( sys.ea = (EntranceAperture *)
	 malloc( sys.n_ea * sizeof(EntranceAperture ) ) ) )
       tf.exit( ExitERR_alloc, "unable to create entrance ap array" );

  {
    SLLNode node;
    EntranceAperture *bin;
    double cum_ea = 0;
    unsigned int id;

    for ( node = sll_head_node( ea_list ),
	      bin = sys.ea,
	      id = 1;
	  node ;
	  node = sll_next_node( ea_list, node ),
	      bin++,
	      id++
	  )
    {
      EntranceAperture *ea = (EntranceAperture *) sll_node_get_data(node );
      *bin = *ea;

      bin->area_min = cum_ea / sys.ea_area;
      cum_ea += bin->area;

      bin->area_max = cum_ea / sys.ea_area;

      if ( bin->id == 0 ) bin->id = id;
    }

    /* free the list of EntranceApertures, but don't free the
       per-aperture data, which contains the aperture specific specs */
    sll_delete( ea_list, free );
  }
}

/* compare a probability against an EntranceAperture structure */
static int
ea_cmp( const void *p, const void *s )
{
  double prob = * (double *) p;
  const EntranceAperture *ea = (EntranceAperture *) s;

  return
    prob < ea->area_min ? -1 :
    prob >= ea->area_max ? 1 :
    0;
}

void
entrance_aperture( Photon *photon, double z )
{
  double prob, x_node_at_ea, y_node_at_ea, distance_from_node, distance_from_ea, radius;
  EntranceAperture *ea;

  /* which entrance aperture to use? */
  prob = ranf( );
  ea = (EntranceAperture *)
    bsearch( &prob, sys.ea, sys.n_ea, sizeof(EntranceAperture), ea_cmp );

  /* if we don't get one, this is a real problem. */
  tf_assert( NULL != ea );

  /* get position on entrance aperture */
  ea->pos( ea->data, &photon->position );

  /* adjust photon direction to account for it hitting away from the
     optical axis */
  {
    /* project the node back along the inital direction vector to the
       plane of the ea */
    distance_from_node =
      ( photon->position.z - sys.node ) / photon->direction.z;
    x_node_at_ea = distance_from_node * photon->direction.x;
    y_node_at_ea = distance_from_node * photon->direction.y;

    /* distance along Z axis times direction vector to get actual distance
       between source and ea */
    distance_from_ea = ( photon->position.z - z ) / photon->direction.z;
    photon->direction.x +=
      ( photon->position.x - x_node_at_ea ) / distance_from_ea;
    photon->direction.y +=
      ( photon->position.y - y_node_at_ea ) / distance_from_ea;
  }

  /* normalize it */
  radius = sqrt( photon->direction.x * photon->direction.x +
		 photon->direction.y * photon->direction.y +
		 photon->direction.z * photon->direction.z );

  photon->direction.x /= radius;
  photon->direction.y /= radius;
  photon->direction.z /= radius;

  photon->eap_id = ea->id;

}



void
ea_cleanup( void )
{
  if ( sys.ea )
  {
    int i;
    for ( i = 0 ; i < sys.n_ea ; i++ )
	sys.ea[i].cleanup(sys.ea[i].data);

    free( sys.ea );
  }
}
