/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2010 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * -->8-->8-->8-->8-- */

#include <math.h>
#include <stdlib.h>

#include <lua.hpp>
#include <lauxlib.h>

#include <mathconst/mathconst.h>

#include <randlib_c/randlib.h>

#include "main.h"
#include "lua_stack.h"
#include "ea.h"

static int	define(lua_State *);
static void	random_pos( void *data, DVector3 *pos );
static void     cleanup( void * data );

luaL_Reg ea_wedge = { "wedge", define };

typedef struct Wedge
{
    double theta_min;
    double theta_max;

    double ri, ro;
    double ri2;
    double dr2;

    double area;
    DVector3 origin;
} Wedge;

static int
define( lua_State *L )
{
    Wedge *data;
    char err_pfx[1024];
    static char *name;

    name = get_lua_cstring( L, "name" );

    /* generate error message prefix */
    sprintf( err_pfx, "ea.wedge `%s'", name );

    if ( NULL == ( data = (Wedge *) malloc( sizeof(Wedge) ) ) )
	luaL_error( L, "%s: error allocating memory", err_pfx );

    /* move opts table to top of Lua stack */
    lua_getfield( L, -1, "opts" );

    data->origin.x  = get_lua_number( L, "x" );
    data->origin.y  = get_lua_number( L, "y" );
    data->origin.z  = get_lua_number( L, "z" );
    data->ri        = get_lua_number( L, "ri" );
    data->ro        = get_lua_number( L, "ro" );
    data->theta_min = get_lua_number( L, "theta_min" );
    data->theta_max = get_lua_number( L, "theta_max" );


    data->ri2 = data->ri * data->ri;
    data->dr2 = data->ro * data->ro - data->ri2;

    /* remove 2 PI degeneracy */
    data->theta_min = fmod( data->theta_min, M_2PI );
    data->theta_max = fmod( data->theta_max, M_2PI );

    /* if the wedge straddles the discontinuity, make it continuous */
    if ( data->theta_min > data->theta_max ) data->theta_max += M_2PI;

    data->area = ( data->theta_max - data->theta_min ) / M_PI * data->dr2 / 100;

    new_entrance_aperture( L, err_pfx,
			   data,
			   get_lua_number( L, "id" ),
			   data->area,
			   random_pos,
			   cleanup
			   );

    free( name );

    return( 0 );
}

static void
random_pos( void *data, DVector3 *pos )
{
  Wedge *wedg = (Wedge *) data;
  double r = sqrt( wedg->ri2 + wedg->dr2 * ranf() );
  double phi = ( wedg->theta_max - wedg->theta_min ) * ranf() + wedg->theta_min;

  pos->x = r * cos( phi ) + wedg->origin.x;
  pos->y = r * sin( phi ) + wedg->origin.y;
  pos->z = wedg->origin.z;
}


static void
cleanup( void *data )
{
  free( data );
}
