/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *       Free Software Foundation, Inc.
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <sys/types.h>
#include <sys/stat.h>
#include <math.h>
#include <limits.h>
#include <float.h>
#include <strings.h>
#include <ctype.h>
#include <unistd.h>

#include <lua.hpp>
#include <lauxlib.h>

#include <tracefctxx/TraceFct.h>
#include <tracefct/exiterrvals.h>
#include <mathconst/mathconst.h>
#include <randlib_c/randlib.h>

#include <suplib/debug.h>
#include <suplib/str.h>
#include <suplib/lists.h>

#include "main.h"
#include "lua_stack.h"
#include "image.h"
#include "misc.h"
#include "sourcepos.h"

using Eigen::Quaterniond;
using Eigen::AngleAxisd;
using Eigen::Vector3d;

#include "pos_gen.h"
#include "units.h"

typedef enum
{
  DF_ERROR = -1,
  DF_PBM,
  DF_PGM,
  DF_RAW_ASCII,
  DF_RAW_FLOAT,
  DF_RAW_DOUBLE,
  DF_FITS
} DataFormat;

static TokListToken
format_type[] =
{
  { "ascii",		DF_RAW_ASCII	},
  { "double",		DF_RAW_DOUBLE	},
  { "fits",		DF_FITS 	},
  { "float",		DF_RAW_FLOAT	},
  { "pbm",		DF_PBM		},
  { "pgm",		DF_PGM		},
};
static TokList format_types = GenTokList(format_type);

typedef enum
{
    FLIP_0  = 0x00,
    FLIP_X  = 0x01,
    FLIP_Y  = 0x10,
    FLIP_XY = 0x11
} FlipWhat;

static TokListToken
flip_type[] =
{
    { "none",   FLIP_0 },
    { "x",	FLIP_X },
    { "y",	FLIP_Y },
    { "xy",	FLIP_XY },
};
static TokList flip_types = GenTokList(flip_type);


typedef struct
{
  double x;		/* x coordinate of pixel */
  double y;		/* y coordinate of pixel */
  double p_min;		/* minimum scaled probability of pixel */
} Pixel;

typedef struct
{
  char *	file;		/* the file to read */
  char *	name;		/* the name of the position generator */
  double	clip;		/* minimum value of a pixel */
  DataFormat	fmt;		/* format of the data */
  double	z;		/* z coord of rect */
  Pixel	*	pixel;		/* the data */
  Axis		x;
  Axis		y;
  size_t	npix;		/* number of pixels */

  SourcePosType pos_type;
  SourcePos     pos;          /* position of source */
  AngleAxisd    rotate;

} ImageData;


static int	define(lua_State *L);
static int	init(PositionGeneratorData *data);
static int	generate(void *data, DVector3 *position, double *z);
static void	cleanup(void *data);

luaL_Reg extent_image = { "image", define };

static void	init_Axis( Axis *axis );
static DataFormat	guess_format( const char *image, int *nx, int *ny );
static void	parse_pixsz( lua_State *L,
			     const char *err_pfx,
			     double dist, ut_unit* unit,
			     Axis* x, Axis* y
			     );

static PositionGeneratorSpec spec =
{
  "image",
  init,
  generate,
  cleanup
};

static int
define( lua_State *L )
{
    ImageData *data;
    char *name;
    char err_pfx[1024];

    TraceFct tf( "image" );

    name = get_lua_cstring( L, "name" );

    /* generate error message prefix */
    sprintf( err_pfx, "image `%s'", name );

    /* allocate space for the image information, and initialize it */
    if ( NULL == ( data = (ImageData *) malloc( sizeof(ImageData) ) ) )
	luaL_error(L, "%s: error allocating memory", err_pfx );

    data->pixel	= NULL;
    data->name	= name;
    init_Axis( &data->x );
    init_Axis( &data->y );

    data->z = get_src_pos( L, err_pfx, &(data->pos_type), &(data->pos) );

    data->rotate = AngleAxisd( get_lua_angular( L, err_pfx, "rotate", 0 ),
			       Vector3d::UnitZ());

    if ( data->pos_type != SOURCE_POS_ANGULAR )
	luaL_error( L, "%s: unsupported source position type", err_pfx );


    /**********************************************************************
    Image File Name

    For now, this is required.  later one will be able to generate or
    reuse internal images

    */
    data->file = get_lua_cstring( L, "file" );

    /* generate new error message prefix, include file name */
    sprintf( err_pfx, "image `%s': file `%s':", name, data->file );

    if ( access( data->file, R_OK ) )
	luaL_error(L, "%s: file does not exist or is not readable", err_pfx);

    /**********************************************************************
   Image size (format dependent)

   This need only be specified as an attribute if it's not in the actual
   image file.  We may need this information to guess at the image file
   type.

   Normlly we'd delete the from the optargs list, but it's an error
   to specify it as an attribute as well as in the image header.  We need
   to keep it around so that when processing image headers we can flag
   the error.

    */
    lua_getfield( L, -1, "size" );
    if ( ! lua_isnil( L, -1 ) )
    {
	data->x.n = get_lua_number( L, "x" );
	data->y.n = get_lua_number( L, "y" );
    }
    lua_pop( L, 1 );


    /**********************************************************************
   Image Format
    */

    {
	char *format = get_lua_cstring( L, "format" );
	if ( format )
	{
	    if ( DF_ERROR == (data->fmt = (DataFormat) tokmatch( format, &format_types )) )
		luaL_error( L, "%s: illegal format type: `%s'",
			    err_pfx, format );
	}


	/* didn't find it */
	else
	{
	    if ( DF_ERROR ==
		 ( data->fmt = guess_format( data->file, &data->x.n, &data->y.n ) ) )
		luaL_error( L, "%s: unable to determine image format; please specify",
			    err_pfx );
	}
    }

    /**********************************************************************
   Read Image header for info

    */
    switch( data->fmt )
    {
    case DF_PBM:
    case DF_PGM:
	if ( exists_lua_field( L, "size" ) )
	    luaL_error( L, "%s: illegal attribute (size) for PGM/PBM images",
			err_pfx );
	rd_image_hdr_pxm( L, err_pfx, data->file, &data->x.n, &data->y.n );
	break;

    case DF_FITS:
	if ( exists_lua_field( L, "size" ) )
	    luaL_error( L, "%s: illegal attribute (size) for FITS images",
			err_pfx );
	rd_image_hdr_fits( L, err_pfx, data->file, data->z, &data->x, &data->y );
	break;

    case DF_RAW_ASCII:
    case DF_RAW_FLOAT:
    case DF_RAW_DOUBLE:
	if ( ! exists_lua_field( L, "size" ) )
	    luaL_error( L, "%s: must specify image size for RAW image formats",
			err_pfx );
	break;

    default:
	luaL_error( L, "%s: internal error at %s: %d", err_pfx, __FILE__, __LINE__ );
    }


    if ( data->x.n <= 0 || data->y.n <= 0 )
	luaL_error( L, "%s: image dimensions  (%d, %d) must be positive",
		    err_pfx, data->x.n, data->y.n );

    /**********************************************************************
   Flip Image?
    */
    {
	int flip = get_lua_token( L, "flip", &flip_types );

	/* be careful here about how flipping is done.  the image format
	   (in particular FITS images with RA&Dec) may have already had
	   a hidden flip; a user request to flip the image should flip
	   the flip, not set it */
	if ( flip & FLIP_X ) data->x.flip = 1 - data->x.flip;
	if ( flip & FLIP_Y ) data->y.flip = 1 - data->y.flip;
    }

    /**********************************************************************
   Clipping Value
    */
    data->clip	= get_lua_number( L, "clip" );


    /**********************************************************************
   Pixel sizes (format dependent)

    while PBM, PGM and the RAW formats all must have a pixel size
    specified, allow FITS images to have their pixel sizes over-ridden
    */


    parse_pixsz( L, err_pfx, data->z, raygen::Units::sys().radian(), &data->x, &data->y );

    data->x.pix_delta = data->x.pix_sz;
    data->y.pix_delta = data->y.pix_sz;

    if ( data->x.flip ) data->x.pix_sz *= -1;
    if ( data->y.flip ) data->y.pix_sz *= -1;

    /* check pixel size */
    if ( -DBL_MAX == data->x.pix_sz || -DBL_MAX == data->y.pix_sz )
	luaL_error( L, "%s: pixel dimensions must be specified", err_pfx );

    if ( 0 == data->x.pix_sz || 0 == data->y.pix_sz )
	luaL_error( L, "%s: pixel dimensions (%f,%f) must be non-zero", err_pfx,
		    data->x.pix_sz, data->y.pix_sz);


    /**********************************************************************
   The position in the image corresponding to the specified Angular
   Source Position

   The user may specify either the pixel index (which may be fractional)
   or the world coordinates of the center.  In the latter case,
   the relationship of the world coordinates must also be specified,
   either via the world coordinate of the pixel
   with indices (1,1), or the pixel with world coordinate (0,0).

   the Lua interface ensures that only the correct permutation is passed in.

    */

    if ( exists_lua_field( L, "point_idx" ) )
    {
	/* which pixel is to centered at the specified angular position of the
	   source */
	lua_getfield( L, -1, "point_idx" );
	data->x.pnt_idx = get_lua_number( L, "x" );
	data->y.pnt_idx = get_lua_number( L, "y" );
	lua_pop( L, 1 );
    }

    else if ( exists_lua_field( L, "point_wcs" ) )
    {
	string error( err_pfx );
	ut_system* system = raygen::Units::sys().system();
	ut_unit* radian = raygen::Units::sys().radian();

	// create "pixel" unit for wcs
	ut_unit* xunit = ut_scale( data->x.pix_sz, radian );
	ut_unit* yunit = ut_scale( data->y.pix_sz, radian );

	/* what wcs position is to be at the specified angular position of the
	   source */
	lua_getfield( L, -1, "point_wcs" );
	ut_map_name_to_unit( "pixel", UT_ASCII, xunit );
	data->x.pnt_wcs = get_lua_extent( L, (error + string("point_wcs.x: ")).c_str(), "x", data->z, radian );
	ut_unmap_name_to_unit( system, "pixel", UT_ASCII );

	ut_map_name_to_unit( "pixel", UT_ASCII, yunit );
	data->y.pnt_wcs = get_lua_extent( L, (error + string("point_wcs.x: ")).c_str(), "y", data->z, radian );
	ut_unmap_name_to_unit( system, "pixel", UT_ASCII );

	lua_pop( L, 1 );

	/* WCS coordinate at reference point */
	lua_getfield( L, -1, "ref_wcs" );
	if ( ! lua_isnil( L, -1 ) )
	{
	    ut_map_name_to_unit( "pixel", UT_ASCII, xunit );
	    data->x.ref_wcs = get_lua_extent( L, (error + string("ref_wcs.x: ")).c_str(), "x", data->z, radian );
	    ut_unmap_name_to_unit( system, "pixel", UT_ASCII );

	    ut_map_name_to_unit( "pixel", UT_ASCII, yunit );
	    data->y.ref_wcs = get_lua_extent( L, (error + string("ref_wcs.y: ")).c_str(), "y", data->z, radian );
	    ut_unmap_name_to_unit( system, "pixel", UT_ASCII );
	}
	lua_pop( L, 1 );

	// all done with pixel units
	ut_free( xunit );
	ut_free( yunit );

	/* pixel index of reference point */
	lua_getfield( L, -1, "ref_idx" );
	if ( ! lua_isnil( L, -1 ) )
	{
	    data->x.ref_idx = get_lua_number( L, "x" );
	    data->y.ref_idx = get_lua_number( L, "y" );
	}
	lua_pop( L, 1 );


	if ( -DBL_MAX == data->x.ref_wcs  || -DBL_MAX == data->y.ref_wcs )
	    luaL_error( L, "%s:  must specify both 'ref_wcs' and 'point_wcs'",
			err_pfx );

	if ( -DBL_MAX == data->x.ref_idx  || -DBL_MAX == data->y.ref_idx )
	    luaL_error( L, "%s: must specify both 'ref_idx' and 'point_wcs'",
			err_pfx );


	/* now, calculate what we really want, which is POINT_IDX_X
	   and POINT_IDX_Y */
	data->x.pnt_idx = data->x.ref_idx +
	    ( data->x.pnt_wcs - data->x.ref_wcs ) / data->x.pix_sz;

	data->y.pnt_idx = data->y.ref_idx +
	    ( data->y.pnt_wcs - data->y.ref_wcs ) / data->y.pix_sz;
    }

    /*
      nothing was specified, so just use the center of the image.
      note that the index of the center of the lower left pixel is
      (1,1)
    */
    else
    {
	data->x.ref_idx = 1.0;
	data->x.ref_wcs = 1.0;
	data->x.pnt_idx = (data->x.n + 1) / 2;

	data->y.ref_idx = 1.0;
	data->y.ref_wcs = 1.0;
	data->y.pnt_idx = (data->y.n + 1) / 2;
    }


    /* inform the ray generator of the new position generator available */
    new_pos_gen( &spec, data->name, data );

    return 0;
}

static void
parse_pixsz( lua_State *L,
	     const char *err_pfx,
	     double dist,
	     ut_unit* unit,
	     Axis* x,
	     Axis* y)
{
    int top = lua_gettop( L );

    lua_getfield( L, -1, "pixsz" );

    if ( ! lua_isnil( L, -1 ) )
    {
	// if it's a table, and it has x & y fields, separate extents
	if ( lua_istable( L, -1 ) &&  exists_lua_field( L, "x" ) && exists_lua_field( L, "y" ) ) {
		x->pix_sz = get_lua_extent( L, err_pfx, "x", dist, unit );
		y->pix_sz = get_lua_extent( L, err_pfx, "y", dist, unit );
	    }

	/* otherwise its a single number, which is shared between x & y */
	else
	    y->pix_sz = x->pix_sz = get_lua_extent( L, string( err_pfx) + " pixsz:", dist, unit );
    }

    lua_settop( L, top );

}

static void
dump_Axis( const char*which, Axis *axis )
{
  TraceFct::message( " %s.n: %d\n", which, axis->n );
  TraceFct::message( " %s.pix_sz: %g\n", which, axis->pix_sz );
  TraceFct::message( " %s.pix_delta: %g\n", which, axis->pix_delta );
  TraceFct::message( " %s.flip: %d\n", which, axis->flip );
  TraceFct::message( " %s.ref_idx: %g\n", which, axis->ref_idx );
  TraceFct::message( " %s.ref_wcs: %g\n", which, axis->ref_wcs );
  TraceFct::message( " %s.pnt_idx: %g\n", which, axis->pnt_idx );
  TraceFct::message( " %s.pnt_wcs: %g\n", which, axis->pnt_wcs );
}

static void
dump_ImageData( ImageData *image )
{
  TraceFct::message( "Image %s:\n", image->name );
  TraceFct::message( " file: %s\n", image->file );
  TraceFct::message( " clip: %f\n", image->clip );
  TraceFct::message( " fmt : %d\n", image->fmt );
  TraceFct::message( " z   : %f\n", image->z );
  TraceFct::message( " npix: %d\n", image->npix );
  TraceFct::message( " theta: %f\n", image->pos.angular.theta );
  TraceFct::message( " phi : %f\n", image->pos.angular.phi );
  dump_Axis( "x", &image->x );
  dump_Axis( "y", &image->y );

}


static int
init( PositionGeneratorData *pgdata )
{
  int x, y;
  double *datap, *data, *data_lim;
  char err_pfx[1024];

  ImageData *image = (ImageData *) pgdata->data;

  double p_sum, p_min;
  Pixel *pixel;

  sprintf( err_pfx, "image `%s': file `%s'", image->name, image->file );

  image->npix = image->x.n * image->y.n;

  switch( image->fmt )
  {
  case DF_RAW_ASCII:
    data = rd_image_data_ascii( err_pfx, image->file, image->npix );
    break;

  case DF_RAW_FLOAT:
    data = rd_image_data_float( err_pfx, image->file, image->npix );
    break;

  case DF_RAW_DOUBLE:
    data = rd_image_data_double( err_pfx, image->file, image->npix );
    break;

  case DF_PBM:
  case DF_PGM:
    data = rd_image_data_pxm( err_pfx, image->file );
    break;

  case DF_FITS:
    data = rd_image_data_fits( err_pfx, image->file );
    break;

  default:
    TraceFct tf( "image init" );
    tf_assert( 0 );
    /* NOT REACHED */
  }

  if ( NULL == data )
    return 1;


  /* warn if the pixel size is a bit small. limit is 1/1000 arcsec */
#define TOO_SMALL_PIX (M_PI / ( 180 * 60 * 60 ) * 0.001)

  if ( fabs(image->x.pix_sz) < TOO_SMALL_PIX )
    TraceFct::message( "%s: x pixel size (%e rad) is less than 0.001 arc seconds\n",
		err_pfx, image->x.pix_sz );

  if ( fabs(image->y.pix_sz) < TOO_SMALL_PIX )
    TraceFct::message( "%s: y pixel size (%e rad) is less than 0.001 arc seconds\n",
		err_pfx, image->y.pix_sz );

  /*
    loop through the data, determine the number of pixels above the
    clip limit
   */

  p_sum = 0.0;
  data_lim = data + image->npix;
  for ( datap = data ; datap < data_lim ; datap++ )
    if ( *datap <= image->clip )
      image->npix--;
    else
      p_sum += *datap;

  image->pixel = new (nothrow) Pixel[image->npix];

  if ( NULL == image->pixel )
  {
    TraceFct::message( "%s: error allocating pixel data (%d bytes)\n", err_pfx,
		image->npix * sizeof(Pixel) );
    return 1;
  }

  /*
    scan through the image, storing "useful" pixels into a list,
    along with their positions and probabilities. convert from
    pixel indices to radians; the position is the left or bottom
    of the pixel, not the center.  the index is 1 at the center
    of the pixel.
  */

  datap = data;
  p_min = 0.0;
  pixel = image->pixel;

  for ( y = 1 ; y <= image->y.n ; y++ )
    for ( x = 1 ; x <= image->x.n ; x++, datap++ )
      if (*datap > image->clip)
      {
        pixel->x = ( x - image->x.pnt_idx - 0.5 ) * image->x.pix_sz;
        pixel->y = ( y - image->y.pnt_idx - 0.5 ) * image->y.pix_sz;
        pixel->p_min = p_min;
	if ( dbflag( "dump_pixel" ) )
	    TraceFct::message( "%f\t%f\t%f\n", pixel->x, pixel->y, *datap );
        p_min += *datap / p_sum;
        ++pixel;
      }

  delete[] data;

  /* prior to this, the magnitude of pix_sz referred to the pixel
     size; it's sign gave the orientation of the image in the data.
     Since the latter is now encoded in the image pixel list, pix_sz
     is now just the pixel size */
  image->x.pix_sz = fabs( image->x.pix_sz );
  image->y.pix_sz = fabs( image->y.pix_sz );

  pgdata->max_extent	= sqrt( (double)
				(image->x.n * image->x.n *
				 image->x.pix_sz * image->x.pix_sz +
				 image->y.n * image->y.n *
				 image->y.pix_sz * image->y.pix_sz
				 ) ) / 2.0 + image->pos.angular.theta;
  pgdata->z		= image->z;
  pgdata->factor	= 1.0;

  if ( dbflag( "dump_image" ) )
    dump_ImageData( image );

  return 0;
}

static int
comp( const void *key, const void *data )
{
  return *(double *) key < ((Pixel *)data)->p_min ? -1 :
         *(double *) key > ((Pixel *)data)->p_min ?  1 :
         0;
}

static int
generate( void *data, DVector3 *direction, double *z )
{
  ImageData *image = (ImageData *) data;
  SourcePosAngular *pos = &(image->pos.angular);
  const Pixel *pixel, *hi_bnd;

  // grab a pixel
  double p = ranf();

  bnd_bsearch( &p, image->pixel, image->npix, sizeof(Pixel),
	       (const void **) &pixel, (const void**) &hi_bnd, comp );
  tf_assert5( NULL != pixel );

  double dx = pixel->x + genunf( 0.0, image->x.pix_sz );
  double dy = pixel->y + genunf( 0.0, image->y.pix_sz );

  // dx and dy are in the direction of the source.  since the direction
  // vector is from the source to the telescope, must flip them.

  // positive rotation about the X axis takes Y into Z, so a positive
  // dy makes the direction vector point in a negative Y direction,
  // which is what we want.

  // positive rotation about the Y axis takes Z into X, so a positive
  // dx makes the direction vector point in a *positive* X direction,
  // which is opposite of what we want.

  Vector3d v =
    *pos->pos
    * image->rotate
    * AngleAxisd( dy, Vector3d::UnitX())
    * AngleAxisd( -dx, Vector3d::UnitY())
    * Vector3d::UnitZ();

  direction->x = v.x();
  direction->y = v.y();
  direction->z = v.z();

  *z	       = pos->z;

  return 0;
}

static void
cleanup( void *data )
{
  ImageData *image = (ImageData *) data;
  if ( image->pixel )
    delete[] image->pixel;
  free( image->name );
  free( image->file );
  free ( data );
}


static DataFormat
guess_format( const char *image, int *nx, int *ny )
{
  size_t nc, i;
  char buf[80];
  FILE *file = fopen( image, "rb" );

  if ( NULL == file )
      return DF_ERROR;

  /* read in the first 80 characters */
  nc = fread( buf, 1, 80, file );
  fclose( file );

  if ( 0 == nc )
      return DF_ERROR;

  if ( 0 == strncasecmp( buf, "simple", 6 ) )
    return DF_FITS;

  if ( 0 == strncasecmp( buf, "P1", 2 ) )
    return DF_PBM;

  if ( 0 == strncasecmp( buf, "P2", 2 ) ||
       0 == strncasecmp( buf, "P5", 2) )
    return DF_PGM;

  /* check for legal characters in ASCII */
  for (i = 0;
       i < nc &&
	 (isspace(buf[i]) || isdigit(buf[i]) || buf[i] == '.' )
	 ; i++ )
    ;

  if ( i == nc )
    return DF_RAW_ASCII;

  /* last hope is to use the file size and the (hopefully specified)
     nx and ny */

  {
    double ipart;
    struct stat buf;

    if ( stat( image, &buf ) )
      return DF_ERROR;

    /* if the axis lengths weren't specified, assume a square image */
    if ( *nx == 0 && *ny == 0 )
    {
      if ( 0.0 == modf( sqrt( buf.st_size / sizeof(float) ), &ipart ) )
      {
	*nx = *ny = sqrt( buf.st_size / sizeof(float) );
	return DF_RAW_FLOAT;
      }
      else if ( 0.0 == modf( sqrt( buf.st_size / sizeof(double) ), &ipart ) )
      {
	*nx = *ny = sqrt( buf.st_size / sizeof(float) );
	return DF_RAW_DOUBLE;
      }
    }

    /* if one axis length was specified, but the other wasn't, can't do much here! */
    else if ( *nx == 0 || *ny == 0 )
      return DF_ERROR;

    /* last try; both were specified */
    else if ( 0.0 == modf( buf.st_size / sizeof(float) / *nx / *ny , &ipart ) )
      return DF_RAW_FLOAT;

    else if ( 0.0 == modf( buf.st_size / sizeof(double) / *nx / *ny , &ipart ) )
      return DF_RAW_DOUBLE;

  }

  return DF_ERROR;
}

static void
init_Axis( Axis *axis )
{
  axis->n	= INT_MIN;
  axis->pix_sz  = -DBL_MAX;
  axis->ref_idx = -DBL_MAX;
  axis->ref_wcs = -DBL_MAX;
  axis->pnt_idx = -DBL_MAX;
  axis->pnt_wcs = -DBL_MAX;
  axis->flip    = 0;

}
