/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *       Free Software Foundation, Inc.
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <math.h>
#include <string.h>

#include <lua.hpp>
#include <lauxlib.h>

#include <tracefctxx/TraceFct.h>
#include <tracefct/exiterrvals.h>

#include "main.h"
#include "common.h"

#include "lua_stack.h"
#include "misc.h"
#include "sourcepos.h"

#include "Jitter/Table.h"
#include "Jitter/File.h"

static int	define(lua_State *);

luaL_Reg jitter_file = { "file", define };

static int
define( lua_State *L )
{
    TraceFct tf( "jitter_file.define");
    char err_pfx[1024];

    string name = get_lua_string( L, "name" );

    /* generate error message prefix */
    sprintf( err_pfx, "raygen.jitter.file.define: `%s'", name.c_str() );

    Jitter::TableFile::Base* file;

    string filename = get_lua_string( L, "file" );
    string format = get_lua_string( L, "format" );

    // first see if we're reading in quaternions or Euler angles
    if ( exists_lua_field( L, "quat_col" ) ) {

	Jitter::TableFile::Quaternion* qfile = new Jitter::TableFile::Quaternion(filename, format );
	get_lua_string( L, err_pfx, "quat_col", qfile->quat_col_name );

	file = qfile;
    }
    else if ( 0 == get_luafield( L, "axes" ) ) {

	Jitter::TableFile::Euler* efile = new Jitter::TableFile::Euler(filename, format );

	// get rotation data if present
	for ( lua_Integer i = 1; i <= 3 ; i++ ) {
	    Jitter::EulerAxis* axis = get_rotation_axis( L, err_pfx, i );
	    if ( NULL == axis )
		break;

	    efile->axes.push_back( axis );
	}

	// from if ( get_luafield() )
	lua_pop( L, 1 );
	file = efile;
    }

    else {

	luaL_error( L, "%s: must specify either quat_col or axes", err_pfx );
    }

    get_lua_string( L, err_pfx, "time_col", file->time_col_name );

    Jitter::Table* jtbl = new Jitter::Table( name );

    get_lua_boolean( L, err_pfx, "rotation_is_additive", jtbl->rotation_is_additive );

    if ( 0 == get_luafield( L, "optical_axis" ) ) {
	string pfx = string(err_pfx) + ".optical_axis";
	get_quat_from_angle_axes( L, pfx, jtbl->optical_axis );
    }

    if ( 0 == get_luafield( L, "input_origin" ) ) {
	string pfx = string(err_pfx) + ".input_origin" ;
	get_quat_from_angle_axes( L, pfx, jtbl->input_origin );
    }

    if ( 0 == get_luafield( L, "coord_xfrm" ) ) {
	string pfx = string(err_pfx) + ".coord_xfrm" ;
	get_quat_from_mat3d( L, pfx, jtbl->coord_xfrm );
    }

    get_lua_number( L,err_pfx, "extrapolate_max_dt", jtbl->extrapolate_max_dt );

    jtbl->read( file );

    delete file;

    sys.jitter.generators.push_back( jtbl );

    return 0;
}
