/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2010 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *       Free Software Foundation, Inc.
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <numeric>
#include <memory>
#include <cfloat>

#include <tracefctxx/TraceFct.h>
#include <tracefct/exiterrvals.h>
#include <suplib/str.h>

#include "main.h"
#include "common.h"
#include "types.h"

#include "rTime.h"
#include "source.h"

#include "jitter_gen.h"

#include <Eigen/Core>
#include <Eigen/Geometry>

using namespace raygen;

using Eigen::Vector3d;
using Eigen::Quaterniond;
using Eigen::AngleAxisd;

struct AccumulateJitter {

    double time;

    Quaterniond q;

    AccumulateJitter( double time) : time(time), q(1,0,0,0) {};
    void operator() ( Jitter::Base* j ) {

	Quaterniond dq;
	j->jitter(time, dq);

#ifdef HARDWIRE
	q = q * dq;
#else

	q = q * j->coord_xfrm * dq * j->coord_xfrm.inverse();
#endif
    }

    operator Quaterniond() { return q; }

};

void
JitterGenerator::apply( double time, DVector3& direction ) {

    Quaterniond q =
	for_each(generators.begin(), generators.end(), AccumulateJitter(time) );

    jitter = jitter * q;

#ifdef HARDWIRE
    Quaterniond q1( AngleAxisd( M_PI/2, Vector3d( 0, 0, 1 ) ) );
    Quaterniond q2( AngleAxisd( M_PI/2, Vector3d( 0, 1, 0 ) ) );

    Quaterniond r = q2 * q1;

    cerr << "r  = " << r << endl;
    cerr << "ri = " << r.inverse() << endl;

    Vector3d in = r * Vector3d( direction.x, direction.y, direction.z );
    Vector3d v = r.inverse() * jitter * in;

#else

    // apply q to our vector.  convert to Eigen Vector3d for now
    Vector3d v = jitter * _roll * Vector3d(direction.x, direction.y, direction.z);

#endif

    direction.x = v.x();
    direction.y = v.y();
    direction.z = v.z();
}

struct TimeBounds : TimeInterval {

    TimeBounds( double min, double max ) : TimeInterval( min, max ) {}
    TimeBounds() {}

    virtual void operator() ( Jitter::Base* ) { throw string( "NOT SUPPOSED TO HAPPEN" ); }
};

struct TimeUnion : TimeBounds {

    TimeUnion() : TimeBounds( DBL_MAX, -DBL_MAX )  { }

    virtual void operator() ( Jitter::Base* j ) {
	if ( min > j->tvalid.min ) min = j->tvalid.min ;
	if ( max < j->tvalid.max ) max = j->tvalid.max ;
    }
};

struct TimeIntersection : TimeBounds {

    TimeIntersection() : TimeBounds( -DBL_MAX,  DBL_MAX ) {}

    virtual void operator() ( Jitter::Base* j ) {
	if ( min < j->tvalid.min ) min = j->tvalid.min ;
	if ( max > j->tvalid.max ) max = j->tvalid.max ;
    }
};

struct TimeDump : TimeBounds {

    string prefix ;

    TimeDump(const char* prefix) : prefix(prefix)  { }

    virtual void operator() ( Jitter::Base* j ) {
	std::cerr
	    << prefix << "dump tmin = " << j->tvalid.min << "\n"
	    << prefix << "dump tmax = " << j->tvalid.max << "\n";
    }
};

void
JitterGenerator::init( Eigen::Quaterniond& roll  ) {

    // figure out merged time bounds

    TimeBounds bnds;

    // I believe that it's possible to do this more elegantly but I'm not smart enough
    if ( sys.config->get_bool( "jitter.tintersect" ) )
	bnds = for_each( generators.begin(), generators.end(), TimeIntersection() );
    else
	bnds = for_each( generators.begin(), generators.end(), TimeUnion() );

    tvalid = bnds;

    _roll = roll;

}
