/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *       Free Software Foundation, Inc.
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <string>
#include <cmath>

using namespace std;

#include <lua.hpp>
#include <lualib.h>
#include <lauxlib.h>

#include <tracefctxx/TraceFct.h>
#include <tracefct/exiterrvals.h>

#include <udunits2.h>
#include <lua_udunits2/lua_udunits2.h>

#include "units.h"
#include "misc.h"
#include "lua_stack.h"


/*---------------------------------------------------------------------*/

static void
type_error( lua_State *L,
	    const string& err_pfx,
	    const char* name,
	    const char* typename_req ) {

    string pfx = err_pfx + '.' + name;
    const char* typename_actual = luaL_typename( L, -1 );

    luaL_error( L, "%s: wrong type. expected a %s, got a %s", pfx.c_str(),
		typename_req, typename_actual );
}

/*! \brief convert the element at the top of the stack into a physical quantity
 *
 * The element at the top of the stack may be a number (in which case
 * the passed \c unit parameter is used as the default unit) or it may
 * be a list containing, sequentially, a number and either a string or
 * a lua_udunits2 encoded userdata.
 *
 * The input value of \c unit is \e copied and used as the unit if no
 * unit is specified.  \c unit may be \c NULL to indicate that there
 * are no default units.  When non-\c NULL, the output value of \c
 * unit is a new \c udunits2 unit object and must be managed by the
 * caller.
 *
 * The stack is unaltered.
 *
 * \param L[in] pointer to Lua state object
 * \param value[out] numeric part of the quantity.
 * \param unit[in,out] pointer to a \c udunits2 unit object describing the units
 * \return a boolean indicating success or failure
 *
 */
static bool
rg_lua_tophysquant( lua_State *L, double &value, ut_unit*& unit ) {

    bool result = true;
    int top = lua_gettop( L );

    // the stack element may be a number or may be a list with two
    // elements, the first a number, the second either a string
    // containing a unit or a userdata containing the unit as a
    // pointer to a ut_unit structure

    try {
	if ( lua_isnumber( L, -1 ) ) {

	    value = lua_tonumber( L, -1 );
	    if ( NULL != unit )
		unit = ut_clone( unit );
	}

	else if ( lua_istable( L, -1 ) ) {

	    lua_pushinteger( L, 1 );
	    lua_gettable( L, -2 );

	    if ( ! lua_isnumber( L, -1 ) )
		throw 0;

	    value = lua_tonumber( L, -1 );
	    lua_pop(L, 1);
	    lua_pushinteger( L, 2 );
	    lua_gettable( L, -2 );

	    if      ( lua_isstring( L, -1 ) ) {

		unit = raygen::Units::parse( lua_tostring(L, -1 ) );

		if ( NULL == unit )
		    throw 0;
	    }
	    else if ( luaut_is_unit( L, -1 ) ) {

		unit = ut_clone( luaut_tounit( L, -1 ) );

	    }

	    else
		throw 0;
	}

    } catch( ... ) {

	result = false;
    }

    lua_settop( L, top );
    return result;
}


/* get a string from the table at the top of the stack  */
char *
get_lua_cstring (
		lua_State *L,
		const char* name
		)
{
    char *str = NULL;

    lua_getfield( L, -1, name );
    if ( lua_isstring( L, -1 ) )
	str = str_dup(lua_tostring( L, -1 ));
    lua_pop( L, 1 );

    return str;
}

/* grab a string from the table at the top of the stack */
string
get_lua_string (
		lua_State *L,
		const char* name
		)
{
    string str;

    lua_getfield( L, -1, name );
    if ( lua_isstring( L, -1 ) )
	str = lua_tostring( L, -1 );
    lua_pop( L, 1 );

    return str;
}

/* grab a string from the table at the top of the stack */
int
get_lua_string (
		lua_State *L,
		const char* err_pfx,
		const char* name,
		string& value
		)
{
    int ret = 0;

    lua_getfield( L, -1, name );
    if ( lua_isstring( L, -1 ) )
	value = lua_tostring( L, -1 );
    else if ( ! lua_isnil( L, -1 ) )
	type_error( L, err_pfx, name, "string");
    else
	ret = 1;

    lua_pop( L, 1 );

    return ret;
}


/* grab a number from the table at the top of the stack */
double
get_lua_number (
		lua_State *L,
		const char* name
		)
{
    double number;

    lua_getfield( L, -1, name );
    number = lua_tonumber( L, -1 );
    lua_pop( L, 1 );

    return number;
}

/* grab a number from the table at the top of the stack */
int
get_lua_number (
		lua_State *L,
		const char* err_pfx,
		const char* name,
		double& value
		)
{
    int ret = 0;

    lua_getfield( L, -1, name );

    if ( lua_isnumber( L, -1 ) )
	value = lua_tonumber( L, -1 );
    else if ( ! lua_isnil( L, -1 ) )
	type_error( L, err_pfx, name, "number" );
    else
	ret = 1;

    lua_pop( L, 1 );

    return ret;
}

/* grab a boolean from the table at the top of the stack */
bool
get_lua_boolean (
		lua_State *L,
		const char* name
		)
{
    bool state;

    lua_getfield( L, -1, name );
    state = lua_toboolean( L, -1 );
    lua_pop( L, 1 );

    return state;
}

/* grab a boolean from the table at the top of the stack */
int
get_lua_boolean (
		lua_State *L,
		const char* err_pfx,
		const char* name,
		bool& value
		)
{
    int ret = 0;

    lua_getfield( L, -1, name );
    if ( lua_isboolean( L, -1 ) )
	value = lua_toboolean( L, -1 );
    else if ( ! lua_isnil( L, -1 ) )
	type_error( L, err_pfx, name, "boolean" );
    else
	ret = 1;

    lua_pop( L, 1 );

    return ret;
}


int
get_lua_token (
	       lua_State *L,
	       const char* name,
	       TokList *tokens )
{
    const char *str;
    int match;

    lua_getfield( L, -1, name );
    str = lua_tostring( L, -1 );

    match = tokmatch( str, tokens );

    lua_pop( L, 1 );

    return match;
}

/** \brief get an extent from the table at the top of the stack.

   Extract a physical quantity from the top of the stack
   and convert it to a linear or angular extent.

   The stack is unaltered.

   \param L[in] the Lua state object
   \param err_pfx[in] a string to prefix error messages
   \param dist[in] the distance in \e mm to the source from the telescope node
   \param t_unit[in] the requested output unit
   \exeption Lua table element not found
   \exeption Lua error in conversion to \e radians
   \return the value in \e radians

 */
double
get_lua_extent(
  lua_State *L,
  const string& err_pfx,
  double dist,
  ut_unit* tunit
  )
{
    double extent;
    ut_unit* unit = NULL;

    ut_unit* mm = raygen::Units::sys().mm();
    ut_unit* radian = raygen::Units::sys().radian();

    if ( ! rg_lua_tophysquant( L, extent, unit ) )
	luaL_error( L, "%s: error parsing unit specification", err_pfx.c_str() );

    if ( NULL == unit )
	luaL_error( L, "%s: unit must be specified for extent", err_pfx.c_str() );


    // figure out what sort of unit we got
    if ( raygen::Units::convert( unit, radian, extent ) ) {

	// angular unit. if can't convert it directly to requested
	// units, they must be linear
	if ( ! raygen::Units::convert( radian, tunit, extent ) ) {
	    // non-cosmological distance conversion
	    extent = fabs(dist) * tan( extent );

	    if ( ! raygen::Units::convert( mm, tunit, extent ) )
		luaL_error( L, "%s: internal error: requested unit is not linear", err_pfx.c_str() );
	}

    }

    else if ( raygen::Units::convert( unit, mm, extent ) ) {

	// linear unit. if can't convert it directly to requested units,
	// they must be angular
	if ( ! raygen::Units::convert( mm, tunit, extent ) ) {

	    // non-cosmological distance conversion
	    extent = atan2( extent, fabs(dist) );

	    if ( ! raygen::Units::convert( radian, tunit, extent ) )
		luaL_error( L, "%s: internal error: requested unit is not angular", err_pfx.c_str() );

	}
    }
    else {

	luaL_error( L, "%s: extent must be a linear or angular distance measure", err_pfx.c_str() );
    }

    if ( unit != NULL )
	ut_free( unit );

    return extent;
}

/** \brief get an extent from the table at the top of the stack.

   Extract a physical quantity from the table at the top of the stack
   and convert it to a linear or angular extent.

   The stack is unaltered.

   \param L[in] the Lua state object
   \param err_pfx[in] a string to prefix error messages
   \param name[in] the name of the element in the table
   \param dist[in] the distance in \e mm to the source from the telescope node
   \param t_unit[in] the requested output unit
   \exeption Lua table element not found
   \exeption Lua error in conversion to \e radians
   \return the value in \e radians

 */
double
get_lua_extent(
  lua_State *L,
  const char *err_pfx,
  const char *name,
  double dist,
  ut_unit* tunit
  )
{
    string error( string(err_pfx) + ' ' + name + ':' );


    lua_getfield( L, -1, name );

    if ( lua_isnil( L, -1 ) )
	luaL_error( L, "%s: unable to get Lua table element: %s", error.c_str() );

    double extent = get_lua_extent( L, error, dist, tunit );

    lua_pop( L, 1 );
    return extent;
}

bool
exists_lua_field( lua_State *L,
		  const char* name )
{
    int exists;
    lua_getfield( L, -1, name );
    exists = ! lua_isnil( L, -1 );
    lua_pop( L, 1 );
    return exists;
}

/*! \brief get a physical quantity from the table at the top of the stack

  Extract a physical quantity from the table at the top of the stack,
  and convert it to the specified unit.  \c value is untouched if the
  requested table element is not defined.

  The stack is left unaltered.

  \param L[in] the Lua state object
  \param var[in] the name of the field in the table
  \param value[out] the value of the quantity
  \param tunit[in] the unit to which to convert to
  \return true unless the unit conversion fails

 */
static bool
get_lua_physquant( lua_State* L,
			const char* var,
			double& value,
			ut_unit* tunit  ) {

    bool result = true;
    ut_unit* unit = tunit;

    lua_getfield( L, -1, var );

    if (    ! lua_isnil( L, -1 )
	 && ! (    rg_lua_tophysquant( L, value, unit )
		&& raygen::Units::convert( unit, tunit, value  )
	     )
	 )
	result = false;

    ut_free( unit );
    lua_pop( L, 1 );
    return result;
}

/*! \brief get a linear measurement from the table at the top of the stack

  Extracts a named element from the table at the top of the stack and
  converts it to \e mm.

  The stack is left unaltered.

  \param L[in] the Lua state object
  \param err_pfx[in] a string to prefix error messages
  \param var[in] the name of the element in the table
  \param value[in] the default value of the quantity, in \e mm
  \exeption Lua error in conversion to \e mm
  \return the value in \e mm

 */
double
get_lua_linear( lua_State* L, const char * err_pfx, const char *var, double value )
{

    if ( ! get_lua_physquant( L, var, value, raygen::Units::sys().mm() ) )
	luaL_error( L, "%s: %s is not a linear distance",
		    err_pfx, var );

    return value;
}

/*! \brief get an angular measurement from the tabled at the top of the stack

  Extracts a named element from the table at the top of the stack and
  converts it to \e radians.

  The stack is left unaltered.

  \param L[in] the Lua state object
  \param err_pfx[in] a string to prefix error messages
  \param var[in] the name of the element in the table
  \param value[in] the default value of the quantity, in \e radians
  \exeption Lua error in conversion to \e radians
  \return the value in \e radians

 */
double
get_lua_angular( lua_State* L, const char * err_pfx, const char *var, double value )
{

    if ( ! get_lua_physquant( L, var, value, raygen::Units::sys().radian() ) )
	luaL_error( L, "%s: %s is not an angular distance",
		    err_pfx, var );

    return value;
}

/*! \brief get time quantity from the table at the top of the stack

  Extracts a named element from the table at the top of the stack and
  converts it to seconds

  The stack is left unaltered.

  \param L[in] the Lua state object
  \param err_pfx[in] a string to prefix error messages
  \param var[in] the name of the element in the table
  \param value[in] the default value of the quantity, in \e seconds
  \exeption Lua error in conversion to \e seconds
  \return the value in \e seconds

 */
double
get_lua_temporal( lua_State* L, const char * err_pfx, const char *var, double value )
{

    if ( ! get_lua_physquant( L, var, value, raygen::Units::sys().second() ) )
	luaL_error( L, "%s: %s is not in units of time",
		    err_pfx, var );

    return value;
}

// resolve the value of a heirarchical lua variable name (starting at
// the global level) push it on the stack.  upon success the value
// will be on the stack and get_lua_var will return 0.  upon failure
// the stack is left unchanged and get_lua_var will return 1;
int
get_luaglobal( lua_State *L, const string& varname, const string& field) {

    lua_getglobal( L, "_G" );

    int err = get_luafield( L, varname, field );

    if ( ! err )
	lua_replace( L, -2 );
    else
	lua_pop( L, 1 );

    return err;
}

// given a table on the Lua stack and a hierarchical named lua
// variable name descend into the table, until all of the components
// in are resolved.  if the passed field parameter is non-empty,
// varname should resolve to a table and field name is in that table.
//
// upon success the value will be on the stack and get_lua_var will return 0.
//
// upon failure the stack is left unchanged and get_luafield will return 1;

int
get_luafield( lua_State *L, const string& varname, const string& field) {


    int top = lua_gettop( L );	// save top of stack

    string::size_type start = 0;
    string::size_type end;

    // duplicate the table to simplify the logic below
    lua_pushvalue( L, -1 );
    do {

	if ( !lua_istable( L, -1 ) ) {
	    lua_settop( L, top );
	    return 1;
	}

	end = varname.find( '.', start );
	string::size_type n_chars = end == string::npos ? varname.size() - start : end - start;

	string svarname( varname, start, n_chars );

	lua_getfield( L, -1, svarname.c_str() );
	if ( lua_isnil( L, -1 ) ) {
	    lua_settop( L, top );
	    return 1;
	}

	// remove parent table
	lua_replace( L, -2 );
	start = end == string::npos ? end : end + 1;

    } while( start != string::npos );

    // if there's a field name then we've got a table on the stack
    if ( ! field.empty() ) {

	if ( !lua_istable( L, -1 ) ) {
	    lua_settop( L, top );
	    return 1;
	}

	lua_getfield( L, -1, field.c_str() );
	if ( lua_isnil( L, -1 ) ) {
	    lua_settop( L, top );
	    return 1;
	}

	// remove parent table
	lua_replace( L, -2 );
    }

    return 0;
}

