/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *       Free Software Foundation, Inc.
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <math.h>

#include <linklist/linklist.h>
#include <randlib_c/randlib.h>
#include <mathconst/mathconst.h>

#include <tracefctxx/TraceFct.h>
#include <tracefct/exiterrvals.h>

#include "main.h"
#include "common.h"

#include "output.h"
#include "photon.h"

void
photon_loop( void )
{
  TraceFct tf( "photon_loop" );

  // be paranoid: make limit_type local (so it can't change value) and check it
  LimitType limit_type = sys.limit_type;
  tf_assert( limit_type == Limit_TIME || limit_type == Limit_COUNTS );


  for( ;; ) {

    /* first photon in system list is the next to get emitted */
    SLLNode node = sll_detach_head_node( sys.gen );
    tf_assert5( NULL != node );

    FluxGenerator* gen = (FluxGenerator *) sll_node_get_data( node );
    Source* source = gen->source;

    if  ( limit_type == Limit_TIME ) {
	if ( source->time > sys.tstop )
	    break;
    }

    else if ( limit_type == Limit_COUNTS ) {

	/* don't assume sys.limit is an integer */
	if ( sys.counts + 1 > sys.limit )
	    break;
    }

    else
	tf.die( "internal error; unexpected limit_type in loop" );

    sys.time = source->time;

    /* generate and output photon */
    generate_photon( source, gen );

    /* replace photon that's been used */
    sll_node_put_data( node, nextPhoton_source( source ) );
    sll_insert_node( sys.gen, node );
  }

  if ( limit_type == Limit_TIME )
      tf_assert( sys.time <= sys.tstop );

  else if ( limit_type == Limit_COUNTS )
      tf_assert( sys.counts == sys.limit );

  else
      tf.die( "internal error; unexpected limit_type after loop" );

}

void
generate_photon( Source *source, FluxGenerator *gen )
{
  double z;
  Photon photon;

  if ( gen->spec.generate( gen->data.data, &photon.energy ) )
    TraceFct::exit( ExitERR_misc, "%s: error generating energy", gen->name );

  source->pos_gen->spec.generate( source->pos_gen->data.data,
				  &photon.direction, &z );

  photon.time = source->time;

  // add some jitter to the photon
  sys.jitter.apply( photon.time, photon.direction );

  /* generate position at entrance pupil */
  entrance_aperture( &photon, z );
  if ( source->dir_gen )
  {
    size_t iter;

    for( iter = 0;
         iter < source->dir_gen->data.iter_lim &&
           source->dir_gen->spec.check( source->dir_gen->data.data,
					&photon.direction ) ;
         iter++ )
    {
      if ( 0 == (iter % source->dir_gen->data.iter_warn) )
	TraceFct::message( "%s: direction generator iterations = %d",
		    source->dir_gen->name, iter);

      entrance_aperture( &photon, z );
    }

    if ( iter == source->dir_gen->data.iter_lim)
	TraceFct::die(
             "%s: direction generator iterations (%d) exceeded limit",
	     source->dir_gen->name, iter );
  }

  ++sys.counts;
  photon.id = sys.counts;
  output( &photon );
}


/*

  PROCEDURE
  nextPhoton_source - Determine the next photon to be emitted by a source.

  DESCRIPTION
  This routine ensures that there is a valid photon time for a source.
  It should be called only after the first photon in the source's valid
  list has been used.  It determines a new arrival time for this photon,
  and ensures that there is at least one valid photon for the source.

  RETURNS
  It exits upon error.

*/

FluxGenerator *
nextPhoton_source(
  Source *	source		/* the source for which to initialize the
  				   photon arrival times */
)
{
  SLLNode node;
  FluxGenerator *gen;

  TraceFct tf( "nextPhoton_source" );

  /* the first photon time in the list has been used and must be updated */
  node = sll_detach_head_node( source->phot_valid );
  tf_assert( NULL != node );

  /* insert the new photon time in the appropriate list */
  if ( update_photon_time( (FluxGenerator *) sll_node_get_data( node ) ) )
    sll_insert_node( source->phot_invalid, node );
  else
    sll_insert_node( source->phot_valid, node );

  /*
    if there are no more valid photons, re-intialize the photon times
    for this source
   */
  if ( 0 == sll_count( source->phot_valid ) )
    init_photon_times( source );

  /* set source time to time of next photon */
  gen = (FluxGenerator *) sll_head( source->phot_valid );
  source->time = gen->time;

  return gen;
}

/*

  PROCEDURE
  init_photon_times - Initialize the photon times for a source.

  RETURNS
  It exits upon error.

*/
void
init_photon_times(
  Source *	source		/* the source for which to initialize the
  				   photon arrival times */
)
{
  size_t iter = 1;

  /*
    initially all of the photon time structures are in
    the phot_invalid list.
  */
  tf_assert( 0 == sll_count(source->phot_temp) );
  tf_assert( 0 == sll_count(source->phot_valid) );

  do
  {
    SLLNode node;
    SLinkList temp;
    Event *event;

    if ( 0 == (iter % source->extent_iter_warn) )
	TraceFct::message( "%s: extent event iterations = %d", source->name, iter );

    /* swap temp and invalid lists */
    temp = source->phot_temp;
    source->phot_temp = source->phot_invalid;
    source->phot_invalid = temp;

    /* generate times for the photons and move photons which have
      arrival times less than the next extent generator
      event time to the valid list
     */
    while (( node = sll_detach_head_node( source->phot_temp ) ))
      if ( update_photon_time( (FluxGenerator *) sll_node_get_data( node ) ) )
        sll_insert_node( source->phot_invalid, node );
      else
        sll_insert_node( source->phot_valid, node );

    tf_assert( 0 == sll_count(source->phot_temp) );

    /*
       if all of the photons are in the invalid list, that means
       that the extent event needs to be serviced and the whole
       shebang tried again.
     */

    if ( 0 != sll_count( source->phot_valid ) )
      break;

    node = sll_detach_head_node( source->extent_events );
    tf_assert( NULL != node );

    event = (Event *) sll_node_get_data( node );
    sll_insert_head_node( sys.spare_events, node );
    source->time = event->time;
    event->service( event->module, event->time,
		    event->module_data, event->event_data );
    event->cleanup( event );

    sll_utraverse( source->flux_gen, set_flux_gen_time, &source->time );
  } while( iter++ < source->extent_iter_lim );

  if ( iter == source->extent_iter_lim)
    TraceFct::exit( ExitERR_misc, "%s: extent event iterations (%d) exceeded limit",
			       source->name, iter );

  /* set source time to time of first photon */
  tf_assert( sll_head( source->phot_valid ) );
  source->time = ((FluxGenerator *) sll_head( source->phot_valid ))->time;
}


/*

  PROCEDURE
  update_photon_time - Update a photon time.

  DESCRIPTION
  This routine updates the time of the next photon emitted by a flux
  generator.

  RETURNS
  It returns zero if the generator is valid, non-zero if it is invalid.
  It exits upon error.

*/
int
update_photon_time(
  FluxGenerator *gen		/* the generator to update */
)
{
  size_t iter = 0;
  Event *event;

  gen->time -= log( ranf( ) ) / ( gen->eff_flux * sys.ea_area );

  /*
     ensure that this photon will arrive before a generator or modifer
     event is scheduled.  since servicing an event implies that either
     the event is rescheduled or deleted, the current event must be
     reacquired for each check.
   */
  while( NULL != (event = (Event *) sll_head( gen->events ))
         && gen->time > event->time && iter++ < gen->data.iter_lim )
  {
    SLLNode node = sll_detach_head_node( gen->events );
    sll_insert_head_node( sys.spare_events, node );

    /* if it's an extent event, this generator is invalid */
    if ( Event_EXTENT == event->type )
      return 1;

    if ( 0 == (iter % gen->data.iter_warn) )
	TraceFct::message( "%s: flux event iterations = %d", gen->name, iter );

    event->service( event->module, event->time,
		    event->module_data, event->event_data );
    event->cleanup( event );

    gen->time = event->time - log( ranf( ) ) / ( gen->eff_flux * sys.ea_area );
  }

  if ( iter == gen->data.iter_lim)
      TraceFct::exit( ExitERR_misc, "%s: flux event iterations (%d) exceeded limit",
	     gen->name, iter );

  return 0;
}
