/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *       Free Software Foundation, Inc.
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <string.h>
#include <math.h>

#include <lua.hpp>
#include <lauxlib.h>

#include <tracefctxx/TraceFct.h>
#include <tracefct/exiterrvals.h>

#include "main.h"
#include "pos_gen.h"
#include "lua_stack.h"
#include "sourcepos.h"

static int	define(lua_State *);
static int	init_angular(PositionGeneratorData *data);
static int	generate_angular(void *data, DVector3 *direction, double *z);
static void	cleanup(void *data);

luaL_Reg extent_point = { "point", define };

typedef struct
{
  SourcePosType pos_type;
  SourcePos     pos;          /* position of source */

  DVector3	direction;	/* direction of point */
  double	z;		/* z coord of point */
} PointData;

static PositionGeneratorSpec spec =
{
  "point",
  init_angular,
  generate_angular,
  cleanup
};

static int
define( lua_State *L )
{
  PointData *data;
  SourcePosAngular *pos;
  char *name;
  char err_pfx[1024];

  TraceFct tf( "point" );

  name = get_lua_cstring( L, "name" );

  /* generate error message prefix */
  sprintf( err_pfx, "point `%s'", name );

  if ( NULL == ( data = (PointData *) malloc( sizeof(PointData) ) ) )
    luaL_error( L, "%s: error allocating memory", err_pfx );

  data->z = get_src_pos( L, err_pfx, &(data->pos_type), &(data->pos) );

  if ( data->pos_type != SOURCE_POS_ANGULAR )
      luaL_error( L, "%s: unsupported source position type", err_pfx );

  pos = &(data->pos.angular);

  data->direction.x = -sin( pos->theta ) * cos( pos->phi );
  data->direction.y = -sin( pos->theta ) * sin( pos->phi );
  data->direction.z =  cos( pos->theta );

  new_pos_gen( &spec, name, data );

  return 0;
}


static int
init_angular( PositionGeneratorData *data )
{
    SourcePosAngular* pos = & (((PointData *) data->data)->pos.angular );

    data->max_extent	= pos->theta;
    data->z		= pos->z;
    data->factor	= 1.0;

    return 0;
}

static int
generate_angular( void *data, DVector3 *direction, double *z )
{
  *z = ((PointData *) data)->z;

  memcpy(direction, &((PointData *) data)->direction, sizeof(DVector3) );

  return 0;
}

static void
cleanup( void *data )
{
  free ( data );
}
