/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *       Free Software Foundation, Inc.
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <iostream>

#include <math.h>
#include <strings.h>

#include <lua.hpp>
#include <lauxlib.h>

#include <fitsio.h>

#include <tracefctxx/TraceFct.h>
#include <tracefct/exiterrvals.h>

#include <suplib/str.h>
#include <suplib/units.h>
#include <suplib/units_linear.h>
#include <suplib/units_angular.h>

#include "main.h"
#include "image.h"

static void get_axis_info(lua_State *L, const char *err_pfx, fitsfile *fptr,
			  double z, int axis_idx, Axis *axis);


void
rd_image_hdr_fits(
		  lua_State *L,
  const char *err_pfx,		/* error message prefix */
  const char *image, 		/* the name of the image file */
  double z, 			/* the position of the source relative
				   to the telescope node */
  Axis *x,
  Axis *y
)
{
  fitsfile *fptr;


  int status = 0;

  if ( fits_open_file( &fptr, image, READONLY, &status ) )
      luaL_error( L, "%s: error opening", err_pfx );

  /* get the size of the image */
  fits_read_key( fptr, TINT, "NAXIS1", &x->n, NULL, &status );
  fits_read_key( fptr, TINT, "NAXIS2", &y->n, NULL, &status );

  if ( status )
  {
      char err_text[128];
      int nstatus = 0;
      fits_close_file( fptr, &nstatus );

      fits_get_errstatus( status, err_text );
      luaL_error( L, "%s: error reading image header: %s\n", err_pfx, err_text);
  }

  get_axis_info( L, err_pfx, fptr, z, 1, x );
  get_axis_info( L, err_pfx, fptr, z, 2, y );

  // determine if we've got RA; flip the axis. sucky way of doing
  // things
  char ctype_val[FLEN_VALUE];
  if ( ! fits_read_key( fptr, TSTRING, "CTYPE1", ctype_val, NULL, &status ) )
  {
    if ( ! strncasecmp( "RA", ctype_val, 2 ) )
      x->flip = 1 - x->flip;
  }

  fits_close_file( fptr, &status );
}


static void
get_axis_info(
	      lua_State *L,
	      const char *err_pfx,		/* an error message prefix string */
	      fitsfile *fptr,		/* the pointer to the FITS file handle */
	      double z,			/* the axial position of the source */
	      int axis_idx,			/* which axis in the FITS file to look up */
	      Axis *axis
	      )
{
  char cunit[FLEN_KEYWORD];
  char cdelt[FLEN_KEYWORD];
  char crpix[FLEN_KEYWORD];
  char crval[FLEN_KEYWORD];
  char ctype[FLEN_KEYWORD];
  char cunit_val[FLEN_VALUE];

  int found_ref = 0;
  int lin_unit = 0;
  int unit;
  int status = 0;

  sprintf( cunit, "CUNIT%d", axis_idx );
  sprintf( cdelt, "CDELT%d", axis_idx );
  sprintf( crpix, "CRPIX%d", axis_idx );
  sprintf( crval, "CRVAL%d", axis_idx );

  /*
    if pixel size is not specified, assume 0.5 arcsec/pixel to
    conform to MARX default
  */
  if ( KEY_NO_EXIST == fits_read_key( fptr, TDOUBLE, cdelt, &axis->pix_sz, NULL, &status ))
  {
    unit = UNIT_arcsecond;
    axis->pix_sz = 0.5;
    status =0;
  }

  /* got a pixel size. now, figure out the units */

  /* if no units are specified, assume degrees */
  else if ( KEY_NO_EXIST ==
	    fits_read_key( fptr, TSTRING, cunit, cunit_val, NULL, &status ))
  {
    status = 0;
    unit = UNIT_degree;
  }

  /* got a unit.  parse it */
  /* is it linear? */
  else  if ( -1 != ( unit = tokmatch( cunit_val, UnitsLinear_def.map ) ) )
    lin_unit = 1;

  /* it had better be angular, then */
  else if ( -1 == ( unit = tokmatch( cunit_val, UnitsAngular_def.map ) ) )
      luaL_error( L, "%s: unknown unit type for keyowrd %s: %s",
		  err_pfx, cunit_val, unit );

  /* check if a reference point was specified */
  if ( !fits_read_key( fptr, TDOUBLE, crpix, &axis->ref_idx, NULL, &status ) )
  {
    if ( ! fits_read_key( fptr, TDOUBLE, crval, &axis->ref_wcs, NULL, &status ) )
      found_ref++;
    else
	luaL_error( L, "%s: found %s, but not %s", err_pfx, crpix, crval );
  }
  status = 0;

  /* if it's a linear unit, use the distance to the source to convert
     it to an angular one */
  if ( lin_unit )
  {
    double cvt = units_cvt( &UnitsLinear_def, unit, UNIT_millimeter );
    axis->pix_sz = atan2( axis->pix_sz * cvt, fabs(z) );
    if ( found_ref )
      axis->ref_wcs = atan2( axis->ref_wcs * cvt, fabs(z) );
  }
  else
  {
    double cvt = units_cvt( &UnitsAngular_def, unit, UNIT_radian );
    status = 0;
    axis->pix_sz *= cvt;
    if ( found_ref )
      axis->ref_wcs *= cvt;
  }

}


double *
rd_image_data_fits(
  const char *err_pfx,		/* error prefix */
  const char *image 		/* the name of the image file */
  )
{
  fitsfile *fptr;

  int nx, ny;

  size_t npix;
  int status = 0;
  int anynul;
  double *data = NULL;

  TraceFct tf( "rd_image_fits_data" );

  if ( fits_open_file( &fptr, image, READONLY, &status ) )
  {
    tf.message( "%s: error opening\n", err_pfx );
    goto error;
  }

  /* get the size of the image */
  fits_read_key( fptr, TINT, "NAXIS1", &nx, NULL, &status );
  fits_read_key( fptr, TINT, "NAXIS2", &ny, NULL, &status );

  if ( status )
  {
    tf.message( "%s: error reading image size\n", err_pfx );
    goto error;
  }

  npix = nx * ny;

  data = new (nothrow) double[npix];

  if ( NULL == data )
  {
    tf.message( "%s: error allocating image data (%d bytes)\n",
    		 err_pfx, npix * sizeof(double) );
    goto error;
  }

  if ( fits_read_img( fptr, TDOUBLE, 1, npix, NULL, data,
		      &anynul, &status ) )
  {
    tf.message( "%s: error allocating image data (%d bytes)\n",
    		 err_pfx, npix * sizeof(double) );
    goto error;
  }

  fits_close_file( fptr, &status );
  return data;

 error:
  if ( data )
    delete[] data;

  if ( fptr )
  {
    status = 0;
    fits_close_file( fptr, &status );
  }

  if ( status )
  {
    char err_text[128];

    fits_get_errstatus( status, err_text );
    tf.message( "%s: %s\n", err_pfx, err_text );
  }

  return NULL;
}
