/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *       Free Software Foundation, Inc.
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include <lua.hpp>
#include <lauxlib.h>

#include <tracefctxx/TraceFct.h>
#include <suplib/str.h>
#include <suplib/io.h>

#include "main.h"
#include "image.h"

/* rd_pbm support routines */

static char *	fget_rec_buf = NULL;
static char *	fget_rec_ptr;
static FILE *	fget_rec_file;
static void *	fget_rec_rec;

static void *
init_fget_rec_buf( FILE *file, void *rec )
{
  fget_rec_file = file;
  fget_rec_rec = rec;
  return
    fget_rec_ptr = fget_rec_buf = fget_rec( fget_rec_file, fget_rec_rec );
}

static int
next_char( void )
{
  if ( '\0' == *fget_rec_ptr && NULL ==
       ( fget_rec_ptr = fget_rec_buf =
	 rd_line( fget_rec_file, fget_rec_rec ) ) )
      return EOF;
  else
    return *fget_rec_ptr++;
}

static int
rd_char( void )
{
  int c;

  do
  {
    if ( EOF == ( c = next_char( ) ) )
      return EOF;
  } while ( isspace( c ) );

  return c;

}

static long int
rd_posint( void )
{
  char *ptr;
  long int c = rd_char( );
  if ( EOF == c )
    return -1;

  fget_rec_ptr--;
  errno = 0;
  c = strtol( fget_rec_ptr, &ptr, 10 );
  if ( errno || c < 0 || ( 0 == c && ( *ptr && !isspace( *ptr ) )) )
    return -2;

  fget_rec_ptr = ptr;

  return c;
}


static double *
read_pbm( const char *err_pfx )
{
  double *datap, *data_lim;
  int c;
  long int npix, nx, ny;

  TraceFct tf( "read_pbm" );

  /* read width */
  if ( ( nx = rd_posint( ) ) < 0 )
  {
    if ( -1 == nx )
      tf.message( "%s: unexpected EOF searching for width\n", err_pfx );

    if ( -2 == nx )
      tf.message( "%s: garbage characters in width\n", err_pfx );

    return NULL;
  }

  /* read height */
  if ( ( ny = rd_posint( ) ) < 0 )
  {
    if ( -1 == ny )
      tf.message( "%s: unexpected EOF searching for height\n", err_pfx );

    if ( -2 == ny )
      tf.message( "%s: garbage characters in height\n", err_pfx );

    return NULL;
  }

  npix = nx * ny;
  double* data = new (nothrow) double[npix];
  if ( NULL == data )
  {
    tf.message( "%s: error allocating image data (%d bytes)\n",
		err_pfx, npix * sizeof(double) );
    return NULL;
  }

  /* read image */
  datap = data;
  data_lim = data + npix;
  while ( datap < data_lim )
  {
    do
    {
      c = rd_char( );
      if ( EOF == c )
      {
	tf.message( "%s: unexpected EOF reading pbm bits\n", err_pfx );
	return NULL;
      }
    } while ( isspace( c ) );

    if ( '0' != c && '1' != c )
    {
      tf.message( "%s: garbage characters in bits in pbm file\n", err_pfx );
      return NULL;
    }
    *datap++ = c == '0' ? 0.0 : 1.0;
  }

  return data;
}

/*
  read the common parts of a pgm, pbm header
*/
void
rd_image_hdr_pxm(
  lua_State *L,
  const char *err_pfx,		/* error message prefix */
  const char *image, 		/* the name of the image file */
  int *nx,
  int *ny
)
{
  void *rec = NULL;
  FILE *file = NULL;
  char magic[3];

#undef ERROR
#define ERROR \
  { if ( rec ) fget_rec_delete( rec ); if ( file ) fclose( file ); }


  if ( NULL == ( file = fopen( image, "r" ) ) )
      luaL_error( L, "%s: error opening image file `%s'\n", err_pfx );

  if ( NULL == (rec = fget_rec_new( 8192, 1024 )) )
  {
    ERROR;
    luaL_error( L, "%s: internal error allocating input buffer\n", err_pfx );
  }

  /* read magic number of input file */
  if ( 2 != fread( magic, 1, 2, file ) )
  {
      ERROR;
      luaL_error( L, "%s: error reading magic number\n", err_pfx );
  }
  magic[2] = '\0';

  /* initiate reading */
  if ( NULL == ( init_fget_rec_buf( file, rec ) ) )
  {
    ERROR;
    luaL_error( L, "%s: internal error setting up input routine", err_pfx );
  }

  /* read width */
  if ( ( *nx = (int) rd_posint( ) ) < 0 )
  {
    if ( -1 == *nx )
    {
      ERROR;
      luaL_error( L, "%s: unexpected EOF searching for width\n", err_pfx );
    }

    if ( -2 == *nx )
    {
      ERROR;
      luaL_error( L, "%s: garbage characters in width\n", err_pfx );
    }
  }

  /* read height */
  if ( ( *ny = (int) rd_posint( ) ) < 0 )
  {
    if ( -1 == *ny )
    {
      ERROR;
      luaL_error( L, "%s: unexpected EOF searching for height\n", err_pfx );
    }

    if ( -2 == *ny )
    {
      ERROR;
      luaL_error( L, "%s: garbage characters in height\n", err_pfx );
    }
  }

  fget_rec_delete( rec );
  fclose( file );
}

static double *
read_pgm( const char *err_pfx, FILE *file, char *magic )
{
  double *data, *datap, *data_lim;
  long int c;
  size_t npix;
  long int nx, ny;

  TraceFct tf( "read_pgm" );

  /* read width */
  if ( ( nx = rd_posint( ) ) < 0 )
  {
    if ( -1 == nx )
	tf.message( "%s: unexpected EOF searching for width\n", err_pfx );

    if ( -2 == nx )
	tf.message( "%s: garbage characters in width\n", err_pfx );

    return NULL;
  }

  /* read height */
  if ( ( ny = rd_posint( ) ) < 0 )
  {
    if ( -1 == ny )
	tf.message( "%s: unexpected EOF searching for height\n", err_pfx );

    if ( -2 == ny )
	tf.message( "%s: garbage characters in height\n", err_pfx );

    return NULL;
  }

  /* read maximum value possible */
  if ( ( c = rd_posint( )) < 0 )
  {
    if ( -1 == c )
      tf.message( "%s: unexpected EOF reading maximum value\n", err_pfx );

    if ( -2 == c )
      tf.message( "%s: garbage characters in maximum value\n", err_pfx );

    return NULL;
  }

  npix = nx * ny;
  if ( NULL == ( data = (double *) malloc( npix * sizeof(double) )) )
  {
    tf.message( "%s, error allocating image data (%d bytes)\n", err_pfx,
    		 npix * sizeof(double) );
    return NULL;
  }

  /* read image */
  datap = data;
  data_lim = data + npix;

  /* ascii */
  if ( 0 == strcmp( magic, "P2" ) )
  {
    for ( ; datap < data_lim ; datap++ )
    {
	if ( ( *datap = (double) rd_posint( ) ) < 0 )
      {
	if ( -1 == *datap )
	  tf.message( "%s: unexpected EOF reading pixel values\n", err_pfx );

	if ( -2 == *datap )
	  tf.message( "%s: garbage characters in pixel values\n", err_pfx );

	return NULL;
      }
    }
  }

  /* raw */
  else
  {
    for ( ; datap < data_lim ; datap++ )
    {
      char c;
      if ( 1 != fread( &c, 1, 1, file ) )
      {
	  tf.message( "%s: unexpected EOF reading pixel values\n", err_pfx );
	  return NULL;
      }
      *datap = c;
    }
  }

  return data;
}


double *
rd_image_data_pxm(
  const char *err_pfx,
  const char *image
  )
{
#undef ERROR
#define ERROR			\
  if ( rec ) fget_rec_delete( rec );	\
  if ( file ) fclose( file );	\
  return NULL

  char magic[3];
  void *rec = NULL;
  double *data;
  FILE *file = NULL;

  TraceFct tf( "rd_image_pxm_data" );

  if ( NULL == ( file = fopen( image, "r" ) ) )
  {
    tf.message( "%s: error opening\n", err_pfx );
    ERROR;
  }

  /* determine magic number of input file */
  if ( 2 != fread( magic, 1, 2, file ) )
  {
    tf.message( "%s: error reading magic number\n", err_pfx );
    ERROR;
  }
  magic[2] = '\0';

  if ( NULL == (rec = fget_rec_new( 8192, 1024 )) )
  {
    tf.message( "%s: error allocating input buffer\n", err_pfx );
    ERROR;
  }

  /* initiate reading */
  if ( NULL == ( init_fget_rec_buf( file, rec ) ) )
  {
    tf.message( "%s: error setting up input routine, err_pfx" );
    ERROR;
  }

  /* PBM */
  if      ( ! strcmp( magic, "P1" ) )
    data = read_pbm( err_pfx );

  /* PGM */
  else if ( ! strcmp( magic, "P2" ) || ! strcmp( magic, "P5") )
    data = read_pgm( err_pfx, file, magic );

  /* Dunno */
  else
  {
    tf.message( "%s: unknown PxM type\n", err_pfx );
    ERROR;
  }

  fget_rec_delete( rec );
  fclose( file );

  return data;
}
