/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *       Free Software Foundation, Inc.
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <math.h>
#include <string.h>

#include <iostream>

#include <lua.hpp>
#include <lauxlib.h>

#include <tracefctxx/TraceFct.h>
#include <tracefct/exiterrvals.h>

#include "main.h"
#include "dist.h"
#include "lua_stack.h"
#include "misc.h"
#include "sourcepos.h"

using Eigen::Quaterniond;
using Eigen::AngleAxisd;
using Eigen::Vector3d;

#include "pos_gen.h"
#include "units.h"

#include "Jitter/jutils.h"

typedef struct
{
  double	half_width;	/* half width of rect */
  double	half_height;	/* half height of rect */

  DVector2	xlim;
  DVector2	ylim;

  ImgDist	idist;

  SourcePosType pos_type;
  SourcePos     pos;          /* position of source */
  AngleAxisd    rotate;
  double	z;	      /* z coord of disk */

} RectData;

static int	define(lua_State *);
static int	init_angular(PositionGeneratorData *data);
static int	generate_angular( void *data, DVector3 *direction, double *z );
static void	cleanup(void *data);

luaL_Reg extent_rect = { "rect", define };

static PositionGeneratorSpec spec =
{
  "rect",
  init_angular,
  generate_angular,
  cleanup
};

static int
define( lua_State *L )
{
  RectData *data;
  const char *name;
  char err_pfx[1024];

  TraceFct tf( "rect" );

  name = get_lua_cstring( L, "name" );

  /* generate error message prefix */
  sprintf( err_pfx, "rect `%s'", name );


  if ( NULL == ( data = (RectData *) malloc( sizeof(RectData) ) ) )
      luaL_error( L, "%s: error allocating memory", err_pfx );

  data->z = get_src_pos( L, err_pfx, &(data->pos_type), &(data->pos) );

  if ( data->pos_type != SOURCE_POS_ANGULAR )
      luaL_error( L, "%s: unsupported source position type", err_pfx );


  ut_unit* radian = raygen::Units::sys().radian();

  /* now that we know the axial location, process the height and
     width, converting into angular units if necessary */
  data->half_height = get_lua_extent( L, err_pfx, "height", data->z, radian );

  if ( data->half_height < 0.0 )
      luaL_error( L, "%s: height must be non-negative", err_pfx );

  data->half_width = get_lua_extent( L, err_pfx, "width", data->z, radian );
  if ( data->half_width < 0.0 )
      luaL_error( L, "%s: width must be non-negative", err_pfx );

  data->half_width	/= 2.0;
  data->half_height	/= 2.0;

  data->rotate = AngleAxisd( get_lua_angular( L, err_pfx, "rotate", 0 ),
			       Vector3d::UnitZ());

  /* determine  intensity distribution */
  get_ImgDist( L, err_pfx, &data->idist, data->z, radian,
	       -data->half_width,  data->half_width,
	       -data->half_height, data->half_height );

  new_pos_gen( &spec, name, data );

  return 0;
}

static int
init_angular( PositionGeneratorData *data )
{
  RectData *rect = (RectData *) data->data;

  data->z		= rect->z;
  data->factor		= 1.0;

  return 0;
}

static int
generate_angular( void *data, DVector3 *direction, double *z )
{
  RectData *rect = (RectData *) data;
  SourcePosAngular *pos = &(rect->pos.angular);
  double dx, dy;

  *z		= pos->z;

  gen_ImgDist( &rect->idist, &dx, &dy );

  // dx and dy are in the direction of the source.  since the direction
  // vector is from the source to the telescope, must flip them.

  // positive rotation about the X axis takes Y into Z, so a positive
  // dy makes the direction vector point in a negative Y direction,
  // which is what we want.

  // positive rotation about the Y axis takes Z into X, so a positive
  // dx makes the direction vector point in a *positive* X direction,
  // which is opposite of what we want.

  Vector3d v =
    *pos->pos
    * rect->rotate
    * AngleAxisd( dy, Vector3d::UnitX())
    * AngleAxisd( -dx, Vector3d::UnitY())
    * Vector3d::UnitZ();

  direction->x = v.x();
  direction->y = v.y();
  direction->z = v.z();

  return 0;
}

static void
cleanup( void *data )
{
  free ( data );
}
