/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *       Free Software Foundation, Inc.
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <math.h>
#include <string.h>
#include <limits.h>

#include <lua.hpp>
#include <linklist/linklist.h>
#include <tracefctxx/TraceFct.h>
#include <tracefct/exiterrvals.h>

#include <suplib/str.h>

#include "main.h"
#include "common.h"
#include "photon.h"

static void	delete_source(void *data);
static int	init_source( void *data );

void
sources_setup( void ) {

    TraceFct tf("sources_setup");

    if ( NULL == ( sys.sources = sll_new( SLL_NULL_CMP ) ) )
	tf.exit( ExitERR_alloc, "unable to initialize source list" );
}

void
sources_init( void )
{
  TraceFct tf("sources_init");

  /* make sure that there's at least one source out there */
  if ( 0 == sll_count( sys.sources ) )
    tf.die( "error: no sources were created" );

  /* ensure that the last source was correctly ended */
  if ( NULL != sys.current_source )
    tf.die( "%s: specification did not end with call to source.end( \"%s\" )",
	     sys.current_source->name, sys.current_source->name );

  /* initialize sources */
  sll_traverse( sys.sources, (SLL_ACTION) init_source );
}

void
sources_cleanup( void )
{
  sll_delete( sys.sources, delete_source );
}


Source *
new_source(
	   lua_State *L,
	   char *name
)
{
  Source *source;

  TraceFct tf( "new_source" );

  if ( NULL == ( source			= (Source *) malloc( sizeof(Source) ))
       ||
       NULL == ( source->phot_temp	= sll_new( flux_gen_time_cmp ) )
       ||
       NULL == ( source->phot_valid	= sll_new( flux_gen_time_cmp ) )
       ||
       NULL == ( source->phot_invalid	= sll_new( flux_gen_time_cmp ) )
       ||
       NULL == ( source->extent_events	= sll_new( event_cmp ) )
       ||
       NULL == ( source->flux_gen	= sll_new( SLL_NULL_CMP ) )
     )
    tf.exit( ExitERR_alloc, "%s: unable to create source" );

  source->name = str_dup( name );

  source->n_gen			= 0;
  source->pos_gen		= NULL;
  source->dir_gen		= NULL;
  source->extent_iter_lim	= ULONG_MAX;
  source->extent_iter_warn	= ULONG_MAX;

  if ( sll_insert_head( sys.sources, source ) )
    tf.die( "%s: error inserting source into system list", source->name );

  return source;
}

static void
delete_source(
  void *data
)
{
  Source *source = (Source *) data;

  free( source->name );

  if ( source->flux_gen )
    sll_delete( source->flux_gen, delete_flux_gen );

  /* the following lists contain flux generators,
     which have already been deleted by the preceding
     sll_delete() call.  we just need to delete the
     list framework */

  if ( source->phot_valid )
    sll_delete( source->phot_valid, NULL );

  if ( source->phot_valid )
    sll_delete( source->phot_invalid, NULL );

  if ( source->phot_valid )
    sll_delete( source->phot_temp, NULL );

  delete_pos_gen( source->pos_gen );

  delete_dir_gen( source->dir_gen );

  sll_delete( source->extent_events, delete_event );

  free( source );
}


/*

  PROCEDURE
  init_source

  RETURNS
  It exits upon error, else it returns zero.

 */
static int
init_source( void *data )
{
  Source *source = (Source *) data;

  TraceFct tf( "init_source" );

  init_pos_gen( source->pos_gen );

  if ( source->dir_gen )
    source->dir_gen->spec.init( &source->dir_gen->data );

  sll_traverse( source->flux_gen, init_flux_gen );

  init_photon_times( source );

  if ( sll_insert( sys.gen, sll_head( source->phot_valid ) ) )
    tf.exit( ExitERR_alloc, "%s: error attaching to system photon list",
	     source->name );
  return 0;
}
