/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2010 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * -->8-->8-->8-->8-- */

#include <math.h>
#include <float.h>

#include <lua.hpp>
#include <lauxlib.h>

#include <suplib/str.h>

#include "lua_stack.h"
#include "units.h"

#include "sourcepos.h"
using Eigen::Quaterniond;
using Eigen::AngleAxisd;
using Eigen::Vector3d;

using namespace std;


static SourcePosType
get_pos_type( lua_State *L )
{
    int ok;

    int table = lua_gettop( L );

    /* figure out the type of source position */
    lua_getfield( L, table, "theta" );
    lua_getfield( L, table, "phi" );
    ok = ! lua_isnil( L, -1) || ! lua_isnil( L, -2 );
    lua_pop( L, 2 );

    if ( ok )
	return SOURCE_POS_ANGULAR;

    /* figure out the type of source position */
    lua_getfield( L, table, "x" );
    lua_getfield( L, table, "y" );
    ok = ! lua_isnil( L, -1) || ! lua_isnil( L, -2);
    lua_pop( L, 2 );

    if ( ok )
	return SOURCE_POS_LINEAR;

    return SOURCE_POS_UNKNOWN;
}

/* assumes there's a table on the top of the Lua stack with an entry
   of "position" which is a table that contains all of the info we need */

double
get_src_pos( lua_State* L,
	     const char * err_pfx, SourcePosType* type, SourcePos *pos )
{
    double src_z;
    int top = lua_gettop( L );

    lua_getfield( L, -1, "position" );

    if ( lua_isnil( L, -1 ) || lua_type( L, -1 ) != LUA_TTABLE )
	luaL_error( L, "%s: no position specified or value not a table", err_pfx );

    *type = get_pos_type( L );

    src_z = get_lua_linear( L, err_pfx, "z", -DBL_MAX );

    switch ( * ((int*) type) )
    {
    case SOURCE_POS_LINEAR:
	pos->linear.x = get_lua_linear( L, err_pfx, "x", 0 );
	pos->linear.y = get_lua_linear( L, err_pfx, "y", 0 );
	pos->linear.z = src_z;

	break;

    default:
	*type = SOURCE_POS_ANGULAR;

    case SOURCE_POS_ANGULAR:
	{
	    SourcePosAngular a;

	    a.z = src_z;
	    // theta & phi are direction TO source
	    a.theta = get_lua_angular( L, err_pfx, "theta", 0 );
	    a.phi   = get_lua_angular( L, err_pfx, "phi", 0 );

	    // construct quaternion to generate direction vectors
	    // The idea is to rotate a Z unit vector by theta
	    // about a Y unit vector which has been rotated around Z by phi
	    Vector3d Yrot( AngleAxisd( a.phi, Vector3d::UnitZ()) * Vector3d::UnitY() );
	    a.pos = new Eigen::Quaterniond( AngleAxisd( -a.theta, Yrot ) );
	    pos->angular = a;
	}
    };

    lua_settop( L, top );

    return src_z;
}

/* assumes there's a table on the top of the Lua stack that
   contains all of the info we need */

using namespace Eigen;



// retrieve a vector from the top of the stack
// it is popped upon success
// a vector is an array with three positional elements.
void
get_vector( lua_State *L,
	  const string& err_pfx,
	  Vector3d& vec  ) {

    // this better be a table
    if ( ! lua_istable( L, -1 ) )
	luaL_error( L, "%s: vec is not a table", err_pfx.c_str() );

    double coords[3];
    for( size_t coord = 1 ; coord <= 3 ; coord++ ) {

	lua_pushinteger( L, coord );
	lua_gettable(L, -2 );

	if ( ! lua_isnumber( L, -1 ) ) {

	    luaL_error( L, "%s: vector coord %d is not a number",
			err_pfx.c_str(), coord );
	}

	coords[coord-1] = lua_tonumber( L, -1 );
	lua_pop( L, 1 );
    }

    vec = Vector3d(coords);
}

// retrieve a vector by name
int
get_vector( lua_State *L,
	  const string& err_pfx,
	  const char* name,
	  Vector3d& vec
	  ) {

    int error = 1;

    int top = lua_gettop( L );

    lua_getfield( L, -1, name );

    if ( ! lua_isnil( L, -1 ) ) {

	string pfx = err_pfx + '.' + name;
	get_vector( L, err_pfx, vec );
	error = 0;
    }

    lua_settop( L, top );

    return error;
}


// retrieve an angle axis from the top of the stack
// it is popped upon success
// an axis is { angle = Lua_Number, axis = vector (see above) }
void
get_angle_axis( lua_State* L,
		const string& err_pfx,
		Eigen::AngleAxisd& axis
		) {

    if ( ! lua_istable( L, -1 ) )
	luaL_error( L, "%s: angle axis must be a table", err_pfx.c_str() );

    double angle = get_lua_angular( L, err_pfx.c_str(), "angle", 0 );

    // get axis vector
    Vector3d vec(0,0,0) ;
    if ( get_vector( L, err_pfx, "axis", vec ) )
	luaL_error( L, "%s: no axis specified", err_pfx.c_str() );

    vec.normalize();
    axis = Eigen::AngleAxisd( angle, vec );

    lua_pop( L, 1 );
}


// retrieve a table of angle axes from the top of the stack
// it is popped upon success
void
get_quat_from_angle_axes( lua_State* L,
			    const string& err_pfx,
			    Quaterniond& q
			    ) {

    int top = lua_gettop( L );

    if ( ! lua_istable( L, -1 ) )
	luaL_error( L, "%s: angle axes must be in a table", err_pfx.c_str() );

    q = Quaterniond::Identity();
    for ( lua_Integer i = 1; i <= 10 ; i++ ) {

	lua_pushinteger( L, i );
	lua_gettable( L, -2 );
	if ( lua_isnil( L, -1 ) )
	    break;

	AngleAxisd axis;

	get_angle_axis( L, err_pfx, axis );

	q = Quaterniond( axis ) * q;
    }

    lua_settop( L, top );
    lua_pop( L, 1 );
}



// get a rotation axis column structure
Jitter::EulerAxis*
get_rotation_axis( lua_State* L,
		   const string& err_pfx,
		   lua_Integer idx
		   ) {

    int top = lua_gettop( L );

    lua_pushinteger( L, idx );
    lua_gettable( L, -2 );
    if ( lua_isnil( L, -1 ) )
    {
	lua_settop( L, top );
	return NULL;
    }

    if ( ! lua_istable( L, -1 ) )
    {
	lua_settop( L, top );
	luaL_error( L, "%s: rotation axis must be a table", err_pfx.c_str() );
    }

    Jitter::EulerAxis* axis = new Jitter::EulerAxis;

    // get rotation axis
    get_vector( L, err_pfx, "axis", axis->axis );
    get_lua_string( L,  err_pfx.c_str(), "col", axis->colname );
    get_lua_number( L,  err_pfx.c_str(), "scale", axis->scale );
    get_lua_boolean( L, err_pfx.c_str(), "autoscale", axis->autoscale );

    lua_settop( L, top );

    return axis;
}


// read in a Matrix3d object as three Vector3d axes
// must be on the stack
// The axes are in an array
// it is popped upon success
int
get_mat3d_from_vec3d( lua_State* L,
                    const string& err_pfx,
                    Matrix3d& mat ) {

    if ( ! lua_istable( L, -1 ) )
	luaL_error( L, "%s: matrix 3d must be a table", err_pfx.c_str() );

    Vector3d x;
    x << mat.block<3,1>(0,0);
    Vector3d y;
    y << mat.block<3,1>(0,0);
    Vector3d z;
    z << mat.block<3,1>(0,0);

    get_vector( L, err_pfx, "x", x );
    get_vector( L, err_pfx, "y", y );
    get_vector( L, err_pfx, "z", z );

    mat << x, y, z;

    lua_pop( L, 1 );
    return 0;
}

// read in a Matrix3d object as three Vector3d axes
int
get_quat_from_mat3d( lua_State* L,
                    const string& err_pfx,
                    Quaterniond& quat ) {

    Matrix3d mat;

    if ( get_mat3d_from_vec3d( L, err_pfx, mat) )
	return 1;

    quat = Quaterniond( mat );

    return 0;
}
