/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *       Free Software Foundation, Inc.
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <math.h>
#include <float.h>

#include <lua.hpp>
#include <lauxlib.h>

#include <mst_rdb/mst_rdb.h>
#include <linklist/linklist.h>

#include <randlib_c/randlib.h>
#include <tracefctxx/TraceFct.h>
#include <tracefct/exiterrvals.h>

#include <suplib/str.h>

#include "main.h"
#include "lua_stack.h"
#include "misc.h"
#include "flux_gen.h"

typedef enum
{
  DF_RDB
} DataFormat;

static TokListToken
format_type[] =
{
  { "rdb",		DF_RDB		},
};
static TokList format_types = GenTokList(format_type);

typedef enum
{
  DU_Photons,
  DU_IntensityErgs,
  DU_IntensityKeV
} DataUnits;

static TokListToken
unit_type[] =
{
  { "ergs/s/cm2/kev",	DU_IntensityErgs },
  { "kev/s/cm2/kev",	DU_IntensityKeV  },
  { "photons/s/cm2",	DU_Photons	 }
};
static TokList unit_types = GenTokList(unit_type);


typedef struct
{
  double e_min;		/* minimum energy in bin */
  double e_max;		/* maximum energy in bin */
  double p_min;		/* minimum scaled probability of pixel */
  double p_max;		/* maximum scaled probability of pixel */
  double binval;	/* input weight of the bin */
} SpectrumBin;

typedef struct
{
  char *e_min;	/* name of column which contains minimum
		   energy of bins */
  char *e_max;	/* name of column which contains maximum
		   energy of bins */
  char *binval;	/* name of column which contains value of bins */
} RDB_Info;

typedef struct
{
  char *	file;
  char *	name;
  DataUnits	units;		/* units of the data */
  DataFormat	format;		/* format of the data file */
  void *	info;		/* format specific information */
  void		(*info_cleanup)(void *);
  SpectrumBin *	bins;		/* the data */
  double	scale;	/* the scale factor */
  long		nbins;		/* number of bins */

} SpectrumData;

static int	define(lua_State *L);
static int	init(FluxGeneratorData *data);
static int	generate(void *data, double *energy);
static void	cleanup(void *data);
static int	comp( const void *key, const void *data );
static void	rd_rdb_spectrum( const char *err_pfx, SpectrumData *
				 spectrum );

static void	rdb_info_cleanup( void *info );

luaL_Reg spectrum_file = { "file", define };

static FluxGeneratorSpec spec =
{
  "file",
  init,
  generate,
  cleanup
};

static int
define( lua_State *L )
{
  SpectrumData *data;
  char *name;
    char err_pfx[1024];

  TraceFct tf( "spectrum file define" );

  name = get_lua_cstring( L, "name" );

  /* generate error message prefix */
  sprintf( err_pfx, "spectrum `%s'", name );

  if ( NULL == ( data = (SpectrumData *) malloc( sizeof(SpectrumData) ) ) )
      luaL_error( L, "%s: error allocating memory", err_pfx );

  data->units = DU_Photons;
  data->scale = 1.0;
  data->format = DF_RDB;
  data->name = name;

  /**********************************************************************
    Spectrum File Name

    For now, this is required.  later one will be able to generate or
    reuse internal spectra

  */
  data->file = get_lua_cstring( L, "file" );


  /* generate new error message prefix, include file name */
  sprintf( err_pfx, "spectrum `%s': file `%s':", name, data->file );

  /**********************************************************************
   Spectrum File Format
  */
  data->format = (DataFormat) get_lua_token( L, "format", &format_types );

  /**********************************************************************
   Spectrum File Units
  */
  data->units = (DataUnits) get_lua_token( L, "units", &unit_types );

  /**********************************************************************
   Scale Factor
  */
  data->scale = get_lua_number( L, "scale" );


  /**********************************************************************
   Format Specific stuff
  */

  switch ( data->format )
  {
  case DF_RDB:
    {
      RDB_Info *info;
      if ( NULL == (data->info = (void *) malloc( sizeof(RDB_Info) )))
	  luaL_error( L, err_pfx, "error allocating file info structure" );
      info = (RDB_Info*) data->info;
      info->e_min = get_lua_cstring( L, "emin" );
      info->e_max = get_lua_cstring( L, "emax" );
      info->binval = get_lua_cstring( L, "flux" );
      data->info_cleanup = rdb_info_cleanup;
      break;
    }

  default:
    tf_assert( 0 );
  }

  new_flux_gen( &spec, name, data );

  return 0;
}

static int
init( FluxGeneratorData *data )
{
  double p_sum, p_min;
  SpectrumBin *binp;
  char err_pfx[1024];

  SpectrumBin *bin;
  SpectrumBin *bin_lim;
  SpectrumData *spectrum = (SpectrumData *) data->data;

  TraceFct tf( "spectrum file init" );

  sprintf( err_pfx, "spectrum `%s': file `%s'", spectrum->name,
	   spectrum->file );


  data->iter_lim = 100;
  data->iter_warn = 10;

  /* read in the data */
  switch ( spectrum->format )
  {
  case DF_RDB:
      rd_rdb_spectrum( err_pfx, spectrum );
      break;

  default:
      tf_assert( 0 );
      break;
  }


  /*
    loop through the data, get the normalized probabilities set up
    and determine the average counts/s/cm2
   */

  bin = spectrum->bins;
  bin_lim = spectrum->bins + spectrum->nbins;
  p_sum = 0.0;
  for ( binp = bin ; binp < bin_lim ; binp++ )
  {
    double e_ave = (binp->e_max + binp->e_min) / 2;
    double e_width = binp->e_max - binp->e_min;

    /* convert data to counts/s/cm2 */
    switch( spectrum->units )
    {
    case DU_Photons:
      break;

    case DU_IntensityErgs:
      /* convert from ergs/s/cm2/keV to counts/s/cm2 */
      binp->binval *= e_width / ( e_ave * 1.6021892e-9 ); /* = ergs/keV */
      break;

    case DU_IntensityKeV:
      /* convert from keV/s/cm2/keV to counts/s/cm2 */
      binp->binval *= e_width / e_ave;
      break;

    default:
      tf_assert( 0 );
    }
    binp->binval *= spectrum->scale;
    p_sum += binp->binval;
  }

  p_min = 0.0;
  data->base_flux = 0;
  for ( binp = bin ; binp < bin_lim ; binp++ )
  {
    binp->p_min = p_min;
    binp->p_max = binp->binval / p_sum + p_min;
    p_min = binp->p_max;

    data->base_flux += binp->binval;
  }

  return 0;
}

static int
generate( void *data, double *energy )
{
  SpectrumData *spectrum = (SpectrumData *) data;
  SpectrumBin *bin;
  double p;

  /* grab a bin */
  p = ranf();

  bin = (SpectrumBin *) bsearch( &p, spectrum->bins, spectrum->nbins,
  			         sizeof(SpectrumBin), comp );
  tf_assert5( NULL != bin );

  *energy = genunf(bin->e_min, bin->e_max);

  return 0;
}

static void
cleanup( void *data )
{
  SpectrumData *spectrum = (SpectrumData *) data;

  if ( ! data )
    return;

  ffree( spectrum->bins );
  ffree( spectrum->file );
  ffree( spectrum->name );
  spectrum->info_cleanup( spectrum->info );
  ffree( spectrum );
}

static void
rdb_info_cleanup( void *info )
{
  if ( info )
  {
    RDB_Info *ptr = (RDB_Info *) info;
    ffree( ptr->e_min );
    ffree( ptr->e_max );
    ffree( ptr->binval );
    ffree( ptr );
  }
}

static void
rd_rdb_spectrum(
  const char *err_pfx,
  SpectrumData * spectrum
)
{
  rdbHeader *rdb = NULL;
  DataColumnMap_st *map = NULL;
  SpectrumBin bin, *bin_p, *bin_lim;
  FILE *fin = NULL;
  RDB_Info *info = (RDB_Info *) spectrum->info;

  static RDBFieldStInfo rdb_fields[] =
  {
    { NULL,	RDB_Num,	offsetof( SpectrumBin, e_min ) },
    { NULL,	RDB_Num,	offsetof( SpectrumBin, e_max ) },
    { NULL,	RDB_Num,	offsetof( SpectrumBin, binval ) }
  };
#define NFIELDS ( sizeof(rdb_fields) / sizeof(RDBFieldStInfo) )

  TraceFct tf( "rd_spectrum_rdb" );

  rdb_fields[0].name = info->e_min;
  rdb_fields[1].name = info->e_max;
  rdb_fields[2].name = info->binval;

  spectrum->bins = NULL;

  if ( NULL == (fin = fopen(spectrum->file, "r" )) )
    tf.die( "%s: error opening file\n", err_pfx );

  if ( NULL == ( rdb = rdb_rd_hdr( fin ) ) )
    tf.die( "%s: error reading rdb header\n", err_pfx );

  map = rdb_map_cols_stst( rdb, NFIELDS, rdb_fields );

  /* determine number of bins */
  spectrum->nbins = 0;
  while( rdb_col_read_st( fin, rdb, map, &bin ) )
    if ( bin.binval > 0.0 )
      spectrum->nbins++;

  if ( ferror( fin ) )
    tf.die( "%s: error reading data\n", err_pfx );

  if ( 0 == spectrum->nbins )
    tf.die( "%s: no bins had a positive value\n", err_pfx );

  if ( NULL ==
       (spectrum->bins = (SpectrumBin *) malloc( spectrum->nbins * sizeof(SpectrumBin) )) )
    tf.die( "%s: unable to allocate bin array of %d elements\n",
		err_pfx, spectrum->nbins );

  /* now, read in bins */
  if ( rdb_rewind( fin, rdb ) )
    tf.die( "%s: unable to rewind. is this a real file?\n", err_pfx );

  bin_p = spectrum->bins;
  bin_lim = bin_p + spectrum->nbins;
  while( bin_p < bin_lim && rdb_col_read_st( fin, rdb, map, bin_p ) )
  {
    if ( bin_p->binval <= 0.0 )
      continue;
    bin_p++;
  }

  if ( bin_p != bin_lim )
    tf.die( "%s: unexpected EOF\n", err_pfx );

  fclose( fin );
  rdb_free_map( map );
  rdb_free_hdr( rdb );
}

static int
comp( const void *key, const void *data )
{
  return *(double *) key < ((SpectrumBin *)data)->p_min ? -1 :
         *(double *) key > ((SpectrumBin *)data)->p_max ?  1 :
         0;
}
