/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *       Free Software Foundation, Inc.
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <cfloat>

#include <mathconst/mathconst.h>
#include <tracefctxx/TraceFct.h>
#include <tracefct/exiterrvals.h>
#include <linklist/linklist.h>
#include <randlib_c/randlib.h>

#include "system.h"

#include "LuaState.h"

#include "main.h"
#include "common.h"
#include "ea.h"
#include "output.h"

// set time bounds
static void
set_time_bounds () {

    TraceFct tf( "set_time_bounds" );

    // sys.tstart is set in gparam.cc:gparam()
    sys.tstop = DBL_MAX;

    // if jitter generators are present, don't force an extrapolation
    if ( ! sys.jitter.empty() ) {

	// if requested tstart is negative, means use
	// jitter start time
	if ( sys.tstart < 0 )
	    sys.tstart = sys.jitter.tmin();

	// make sure that doesn't violate the jitter time boundaries
	if ( sys.tstart < sys.jitter.tmin() )
	    tf.die( "start time (%f) < jitter start time (%f)",
		    sys.tstart, sys.jitter.tmin() );

	if ( sys.tstart > sys.jitter.tmax() )
	    tf.die( "start time (%f) > jitter stop time (%f)",
		    sys.tstart, sys.jitter.tmax() );
    }

    // if the specified limit is time and is positive, use that to set
    // sys.tmax

    if ( sys.limit_type == Limit_TIME) {

	// no jitter? use the limit directly
	if ( sys.jitter.empty() ) {

	    if ( sys.limit <= 0)
		tf.die( "time limit (%f) must be > 0", sys.limit );

	    sys.tstop = sys.tstart + sys.limit;

	}

	// jitter? sys.limit may have special value
	else {

	    // use jitter upper range
	    if ( sys.limit == 0 ) {

		sys.tstop = sys.jitter.tmax();

	    }

	    else {

		if ( sys.limit < 0)
		    tf.die( "time limit (%f) must be >= 0 if jitter is on", sys.limit );

		sys.tstop = sys.tstart + sys.limit;

		if ( sys.tstop > sys.jitter.tmax() )
		    tf.die( "stop time (%f = %f + %f) > jitter stop time (%f)",
			    sys.tstop, sys.tstart, sys.limit,
			    sys.jitter.tmax() );
	    }

	}

	if ( sys.tstart >= sys.tstop )
	    tf.die( "start time (%f) >= stop time (%f)",
		    sys.tstart, sys.tstop );

	tf_assert( sys.tstop != DBL_MAX );

	tf.message( "tstart = %f\ntstop = %f\n", sys.tstart, sys.tstop );
    }

}

void
system_init( void )
{
    TraceFct tf( "system_init" );

    /* initialize random number generator and get to correct block */
    setall( sys.seed1, sys.seed2 );
    for ( size_t n = 0 ; n < sys.block ; n++ )
	initgn(1);

    if ( NULL == ( sys.gen = sll_new( flux_gen_time_cmp ) ) )
	tf.exit( ExitERR_alloc, "unable to initialize flux generators" );

    if ( NULL == ( sys.spare_events = sll_new( SLL_NULL_CMP ) ))
	tf.exit( ExitERR_alloc, "unable to initialize spare event queue" );

    sys.counts = 0;

    sys.roll = Eigen::Quaterniond( 1, 0, 0, 0 );

    sources_setup();
    ea_setup();

    // start up Lua and run raytrace configuration script
    LuaState L = LuaState();

    // set up configuration object
    sys.config = new SysConf( L, "saotrace.config", "raygen" );
    sys.config->set( "logfile", sys.logfile );

    // initialize raygen lua setup
    lua_pushstring( L, sys.cfgfile );
    lua_pushstring( L, sys.cfgpars );
    L.call_module_func( "saotrace.raygen", "init", 2, 0 );

    sys.config->get( "node", &sys.node );

    // this must be done first as it may affect the starting time
    sys.jitter.init( sys.roll );
    set_time_bounds();

    /* initialize the sources and entrance apertures */
    sources_init( );
    ea_init( );

    /* create the output stream */
    output_init( );

    /* write some messages out */
    tf.message( "ea_area = %f\n", sys.ea_area );

    /* if limit is density, convert to counts */
    if ( Limit_DENSITY == sys.limit_type )
    {
	sys.limit = (int) ( sys.limit * sys.ea_area + 0.5);
	sys.limit_type = Limit_COUNTS;
    }

    tf.message( "limit\t= %f\n", sys.limit );
}

void
system_cleanup( void )
{
  TraceFct tf( "system_cleanup" );

  output_cleanup( );

  sources_cleanup( );

  ea_cleanup( );

  sll_delete( sys.gen, SLL_NULL_DELETE );
  sll_delete( sys.spare_events, free );

}

