#include <cstdio>
#include <cstdarg>
#include <cstring>

#include <lua.hpp>

#include <lua_udunits2/lua_udunits2.h>

#include <suplib/str.h>
#include <suplib/units.h>
#include <suplib/units_linear.h>
#include <suplib/units_angular.h>

#include "units.h"

namespace raygen {

    char Units::_errmsg[1024] = {};

    Units::Units() {
	_errmsg[0] = '\0';
	ut_set_error_message_handler(handle_error_message);
	_system = ut_read_xml( NULL );

	if ( NULL == _system )
	    throw Exception( string(_errmsg) );

	if ( NULL == (_radian = ut_get_unit_by_name( _system, "radian" ) ) )
	    throw Exception( string("unable to obtain radian unit") );

	if ( NULL == (_second = ut_get_unit_by_name( _system, "second" ) ) )
	    throw Exception( string("unable to obtain radian unit") );

	if ( NULL == ( _meter = ut_get_unit_by_name( _system, "meter" ) ) )
	    throw Exception( string( "unable to obtain meter unit" ) );
	_mm = ut_scale( .001, _meter );

	// make sure that FITS 'deg' UNITS is recognized
	ut_unit* degree = ut_get_unit_by_name( _system, "degree" );
	ut_map_name_to_unit( "deg", UT_ASCII, degree );
	ut_free( degree );
    }

    Units::~Units() {
	ut_free( _radian );
	ut_free( _mm );
	ut_free( _meter );
	ut_free( _second );
	ut_free_system( _system );
    }

    ut_unit*
    Units::parse( const char* str ) {

	size_t len = strlen( str );
	char* buf = new char[len + 1];
	strcpy( buf, str );

	ut_trim( buf, UT_ASCII );

	ut_unit* unit = ut_parse( sys()._system, buf, UT_ASCII );
	delete[] buf;

	return unit;
    }


    int
    Units::handle_error_message( const char *fmt, va_list ap ) {

	int n = vsnprintf( _errmsg, sizeof( _errmsg), fmt, ap );

	return n;
    }

    bool
    Units::convert( string from_str, ut_unit* to, double& value ) {

	ut_unit* from = parse( from_str );

	if ( NULL == from )
	    return false;

	return convert( from, to, value );
    }

    bool
    Units::convert( ut_unit* from, ut_unit* to, double& value ) {

	cv_converter* converter = ut_get_converter( from, to );
	if ( NULL == converter )
	    return false;

	value = cv_convert_double( converter, value );

	return true;
    }

}

static int
_units_system( lua_State *L ) {

    ut_system* system = raygen::Units::sys().system();
    luaut_push_udata_system(L, system, 0 );

    return 1;

}

luaL_Reg units_system = { "system", _units_system };

int cvt_linear( const char *str, double *value )
{
    UnitVal uv = { 0, UNIT_millimeter };
    int err = units_parse( str, &uv, &UnitsLinear_def );

    if ( err == UNITS_NOSPEC || err == UNITS_OK )
	*value = uv.value * units_cvt( &UnitsLinear_def, uv.unit,
				   UNIT_millimeter );

    else
	return err;


    return 0;
}


int cvt_angular( const char *str, double *value )
{
    UnitVal uv = { 0, UNIT_radian };
    int err = units_parse( str, &uv, &UnitsAngular_def );

    if ( err == UNITS_NOSPEC || err == UNITS_OK )
	*value = uv.value * units_cvt( &UnitsAngular_def, uv.unit,
				       UNIT_radian );
    else
	return err;

    return 0;
}
