#include <string>
#include <iostream>

#include "gtest/gtest.h"
#include "SimpleFits.h"
#include "Jitter/Table.h"
#include "q_utils.h"

class JitterTest : public ::testing::Test {

protected:

    // first, penultimate and ultimate quaternions in the input test file
    Quaterniond q0;
    Quaterniond qn_1;
    Quaterniond qn;

    // first, penultimate and ultimate times in the input test file
    double t0;
    double tn_1;
    double tn;

    size_t tbl_size;
    string table;

    virtual void SetUp() {

	// force exception if env var is not set
	table = string(getenv( "RAYGEN_TEST_JITTER_TABLE" ));

	SimpleFits::Table tbl( table );
	tbl_size = tbl.num_rows();

	double qarr[4];
	double *ptr;
	SimpleFits::Table::Column q_att = tbl.column( "q_att", "q_att" );
	SimpleFits::Table::Column time  = tbl.column( "time", "time" );

	tbl.read_col_double( (ptr=&t0), time, 1, 1 );
	tbl.read_col_double( (ptr=qarr), q_att, 1, 1 );
	q0 = Quaterniond( qarr );

	tbl.read_col_double( (ptr=&tn_1), time, tbl_size-1, 1 );
	tbl.read_col_double( (ptr=qarr), q_att, tbl_size-1, 1 );
	qn_1 = Quaterniond( qarr );

	tbl.read_col_double( (ptr=&tn), time, tbl_size, 1 );
	tbl.read_col_double( (ptr=qarr), q_att, tbl_size, 1 );
	qn = Quaterniond( qarr );

    }

};




TEST_F(JitterTest, read_tbl_quat ) {

    Jitter::Table tbl( "test" );
    tbl.rotation_is_additive = false;

    Jitter::TableFile::Quaternion file( table, "fits" );
    file.rowsize = 20;

    tbl.read( &file );

    ASSERT_EQ( tbl_size, tbl.size() );

    EXPECT_TRUE( IsEqual( q0, tbl.points[0].q) );
    EXPECT_TRUE( IsEqual( qn, tbl.points[tbl.size()-1].q) );

    EXPECT_DOUBLE_EQ( t0, tbl.points[0].time );
    EXPECT_NEAR( tn, tbl.points[tbl.size()-1].time, 1e-7 );
}

TEST_F(JitterTest, read_tbl_comp ) {

    // first read in quaternions to test against
    Jitter::Table quat( "test" );
    quat.rotation_is_additive = false;

    {
	Jitter::TableFile::Quaternion file( table, "fits" );
	file.rowsize = 20;

	quat.read( &file );
    }

    ASSERT_EQ( tbl_size, quat.size() );

    Jitter::Table comp( "test" );
    comp.rotation_is_additive = false;

    // test data are in celestial coords (Right Ascension, Declination); must flip Y axis
    // to match space craft coords
    {
	Jitter::TableFile::Euler file( table, "fits" );

	file.axes.push_back( new Jitter::EulerAxis( "ra",   "ra",   Eigen::Vector3d( 0,  0, 1 ), 1.0 ) );
	file.axes.push_back( new Jitter::EulerAxis( "dec",  "dec",  Eigen::Vector3d( 0, -1, 0 ), 1.0 ) );
	file.axes.push_back( new Jitter::EulerAxis( "roll", "roll", Eigen::Vector3d( 1,  0, 0 ), 1.0 ) );

	comp.read( &file );
    }

    ASSERT_EQ( tbl_size, comp.size() );

    for( size_t idx = 0 ; idx < comp.size() ; idx++ )
	EXPECT_TRUE( IsEqual( quat.points[idx].q, comp.points[idx].q) );
}


TEST_F(JitterTest, interpolate ) {

    // grab RA, DEC and ROLL nom

    SimpleFits::Table fits( table );

    // grab nominal pointing and convert to radians
    double RA_NOM   = fits.read_keyword_double( "RA_NOM" )   * M_PI / 180;
    double DEC_NOM  = fits.read_keyword_double( "DEC_NOM" )  * M_PI / 180;
    double ROLL_NOM = fits.read_keyword_double( "ROLL_NOM" ) * M_PI / 180;

    Vector3d X( 1, 0, 0 );
    Vector3d Y( 0, 1, 0 );
    Vector3d Z( 0, 0, 1 );

    Jitter::Table tbl( "test" );
    tbl.rotation_is_additive = false;

    // optical axis is the nominal pointing direction
    tbl.optical_axis =
	  Quaterniond( AngleAxisd( ROLL_NOM, X ) )
	* Quaterniond( AngleAxisd( DEC_NOM,  Y ) )
	* Quaterniond( AngleAxisd( RA_NOM,   Z ) );

    // make sure this ain't null
    ASSERT_TRUE( IsNotEqual( Quaterniond( 1, 0, 0, 0 ), tbl.optical_axis ) );

    // reading quaternions in, but using celestial coords for optical
    // axis; must convert that to quaternions' coord
    Matrix3d xfrm;
    xfrm << 1,  0, 0,
	    0, -1, 0,
	    0,  0, 1;

    tbl.optical_axis.vec() = xfrm * tbl.optical_axis.vec();

    Jitter::TableFile::Quaternion file( table, "fits" );
    tbl.read( &file );

    // first quaternion should be rotation from optical_axis to q0
    EXPECT_TRUE( IsEqual( tbl.optical_axis.inverse() * q0, tbl.points[0].q ) );

    // jitter is the cumulative rotation, dj is the delta from the last call to jitter();
    Quaterniond jitter( 1, 0, 0, 0);
    Quaterniond dj;


    if ( getenv( "RAYGEN_TEST_DUMP_INTERPOLATE" ) ) {
	    std::cerr
		<< "time\t"
		<< "v0_x\t"
		<< "v0_y\t"
		<< "v0_z\t"
		<< "v1_x\t"
		<< "v1_y\t"
		<< "v1_z\n"
		<< "N\t"
		<< "N\t"
		<< "N\t"
		<< "N\t"
		<< "N\t"
		<< "N\t"
		<< "N\n"
		;
    }

    // move through quaternions in fractions of jitter timesteps to see
    // if the error between the additive rotations is small
    for ( size_t idx = 0 ; idx < tbl.size()-1 ; idx++ ) {

	double dt[3] = { 0, 0.25, 0.8 };

	double tp = tbl.points[idx+1].time - tbl.points[idx].time;

	for ( size_t tidx = 0 ; tidx < sizeof(dt) / sizeof(double) ; tidx++ ) {

	    double time = tbl.points[idx].time + dt[tidx] * tp;
	    tbl.jitter( time , dj );

	    Quaterniond q = tbl.points[idx+1].q.slerp( dt[tidx], tbl.points[idx].q);

	    jitter = jitter * dj;

	    if ( getenv( "RAYGEN_TEST_DUMP_INTERPOLATE" ) ) {
		Vector3d v0 = q * X;
		Vector3d v1 = jitter * X;

		std::cerr
		    << time << "\t"
		    << v0.x() << "\t"
		    << v0.y() << "\t"
		    << v0.z() << "\t"
		    << v1.x() << "\t"
		    << v1.y() << "\t"
		    << v1.z() << "\n"
		    ;
	    }


	    EXPECT_TRUE( IsEqual( q , jitter, 5e-5 ) );
	}

    }



}

TEST_F(JitterTest, out_of_bounds ) {

    Jitter::Table tbl( "test" );
    tbl.rotation_is_additive = false;

    Jitter::TableFile::Quaternion file( table, "fits" );

    tbl.read( &file );

    {
	Quaterniond q;

	EXPECT_THROW(tbl.jitter( tbl.tvalid.min-1, q ), Exception  );
    }

    {
	Quaterniond q;

	EXPECT_THROW( tbl.jitter( tbl.tvalid.max, q ), Exception  );
    }

}
