#include <string>
#include <iostream>

#include <lua.hpp>
#include <lauxlib.h>

#include "gtest/gtest.h"
#include "t_utils.h"

#include "lua_stack.h"

using namespace std;

static int
panic( lua_State *L ) {

    string errstr(  lua_tostring( L, -1 ) );
    lua_pop( L, -1 );
    throw errstr;
}


class StackTest : public testing::Test {

protected:

    lua_State *L;

    virtual void SetUp() {

	L = luaL_newstate();
	luaL_openlibs( L ) ;
    }

    virtual void TearDown() {

	lua_close( L );

    }
};

TEST_F( StackTest, get_luaglobal ) {

    luaL_dostring( L, "raygen = { config = { x = 5 } }" );

    ASSERT_EQ( 0, get_luaglobal( L, "raygen" ) );
    ASSERT_TRUE( lua_istable(L, -1 ) );
    lua_getfield( L, -1, "config" );
    ASSERT_FALSE( lua_isnil( L, -1 ) );
    lua_pop( L, 2 );

    ASSERT_EQ( 0, get_luaglobal( L, "raygen.config" ) );
    ASSERT_TRUE( lua_istable(L, -1 ) );
    lua_getfield( L, -1, "x" );
    ASSERT_FALSE( lua_isnil( L, -1 ) );
    lua_pop( L, 2 );

    ASSERT_EQ( 0, get_luaglobal( L, "raygen.config.x" ) );
    ASSERT_EQ( 5, lua_tointeger( L, -1 ) );
    lua_pop( L, 1 );

    ASSERT_EQ( 0, get_luaglobal( L, "raygen.config", "x" ) );
    ASSERT_EQ( 5, lua_tointeger( L, -1 ) );
    lua_pop( L, 1 );


    ASSERT_EQ( 1, get_luaglobal( L, "xy" ) );
    ASSERT_EQ( 1, get_luaglobal( L, "raygen.xy" ) );
    ASSERT_EQ( 1, get_luaglobal( L, "raygen.config.xy" ) );
}

TEST_F( StackTest, get_lua_boolean ) {

    lua_CFunction _panic;
    bool value = false;
    int nil;

    lua_newtable( L );

    // a boolean
    lua_pushboolean( L, true );
    lua_setfield( L, -2, "name" );

    _panic = lua_atpanic( L, panic );
    ASSERT_NO_THROW( nil = get_lua_boolean( L, "pfx", "name", value ) );
    ASSERT_EQ( 0, nil );
    ASSERT_TRUE( value );
    lua_atpanic( L, _panic );

    ASSERT_TRUE( get_lua_boolean( L, "name" ) );

    // a nil
    lua_pushnil( L );
    lua_setfield( L, -2, "name" );

    _panic = lua_atpanic( L, panic );
    ASSERT_NO_THROW( nil = get_lua_boolean( L, "pfx", "name", value ) );
    ASSERT_EQ( 1, nil );
    lua_atpanic( L, _panic );


    // Panic test is at end to avoid unknown state of Lua Stack
    lua_pushstring( L, "value" );
    lua_setfield( L, -2, "name" );

    lua_atpanic( L, panic );
    ASSERT_THROW( get_lua_boolean( L, "pfx", "name", value ), string );
    lua_atpanic( L, _panic );
}

TEST_F( StackTest, get_lua_number ) {

    lua_CFunction _panic;
    double value = -1;
    int nil;

    lua_newtable( L );

    // a boolean
    lua_pushnumber( L, 1 );
    lua_setfield( L, -2, "name" );

    _panic = lua_atpanic( L, panic );
    ASSERT_NO_THROW( nil = get_lua_number( L, "pfx", "name", value ) );
    ASSERT_EQ( 0, nil );
    ASSERT_EQ( 1, value );
    lua_atpanic( L, _panic );

    ASSERT_EQ( 1, get_lua_number( L, "name" ) );

    // a nil
    lua_pushnil( L );
    lua_setfield( L, -2, "name" );

    _panic = lua_atpanic( L, panic );
    ASSERT_NO_THROW( nil = get_lua_number( L, "pfx", "name", value ) );
    ASSERT_EQ( 1, nil );
    lua_atpanic( L, _panic );


    // Panic test is at end to avoid unknown state of Lua Stack
    lua_pushstring( L, "value" );
    lua_setfield( L, -2, "name" );

    lua_atpanic( L, panic );
    ASSERT_THROW( get_lua_number( L, "pfx", "name", value ), string );
    lua_atpanic( L, _panic );
}

TEST_F( StackTest, get_lua_string ) {

    lua_CFunction _panic;
    string value( "default" );
    int nil;

    lua_newtable( L );

    // a boolean
    lua_pushstring( L, "not default" );
    lua_setfield( L, -2, "name" );

    _panic = lua_atpanic( L, panic );
    ASSERT_NO_THROW( nil = get_lua_string( L, "pfx", "name", value ) );
    ASSERT_EQ( 0, nil );
    ASSERT_EQ( "not default", value );
    lua_atpanic( L, _panic );

    ASSERT_EQ( "not default", get_lua_string( L, "name" ) );

    // a nil
    lua_pushnil( L );
    lua_setfield( L, -2, "name" );

    _panic = lua_atpanic( L, panic );
    ASSERT_NO_THROW( nil = get_lua_string( L, "pfx", "name", value ) );
    ASSERT_EQ( 1, nil );
    lua_atpanic( L, _panic );

    // Panic test is at end to avoid unknown state of Lua Stack
    lua_pushboolean( L, false );
    lua_setfield( L, -2, "name" );

    lua_atpanic( L, panic );
    ASSERT_THROW( get_lua_string( L, "pfx", "name", value ), string );
    lua_atpanic( L, _panic );
}
