#include <string>
#include <iostream>

#include <cfloat>

#include <lua.hpp>
#include <lauxlib.h>

#include <lua_udunits2/lua_udunits2.h>

#include "gtest/gtest.h"

#include "lua_stack.h"
#include "units.h"

static int
panic( lua_State *L ) {

    std::string errstr(  lua_tostring( L, -1 ) );
    lua_pop( L, -1 );
    cerr << errstr << endl;
    throw errstr;
}

static int
panic_noerr( lua_State *L ) {

    std::string errstr(  lua_tostring( L, -1 ) );
    lua_pop( L, -1 );
    throw errstr;
}


class PhysQuant : public testing::Test {

protected:

    lua_State *L;

    virtual void SetUp() {

	L = luaL_newstate();
	luaL_openlibs( L ) ;
	luaL_dostring( L, "require 'udunits2'" );
	lua_atpanic( L, panic );

	lua_newtable( L );

    }

    virtual void TearDown() {

	lua_close( L );

    }
};

TEST_F( PhysQuant, get_lua_linear_string ) {

    lua_newtable( L );	// physquant table
    int table = lua_gettop( L );

    lua_pushinteger( L, 1 );	// key
    lua_pushnumber( L, 2 );	// value
    lua_settable( L, table );

    lua_pushinteger( L, 2 );	// key
    lua_pushstring( L, "m" );	// value
    lua_settable( L, table );

    lua_setfield( L, -2, "name" );

    EXPECT_FLOAT_EQ( 2000, get_lua_linear( L, "pfx", "name", 0 ) );

}

TEST_F( PhysQuant, get_lua_linear_unit ) {

    lua_newtable( L );	// physquant table
    int table = lua_gettop( L );

    lua_pushinteger( L, 1 );	// key
    lua_pushnumber( L, 2 );	// value
    lua_settable( L, table );

    ut_unit* meter = raygen::Units::parse( "meter" );

    lua_pushinteger( L, 2 );	// key
    luaut_push_udata_unit(L, meter, 0);  // value
    lua_settable( L, table );

    lua_setfield( L, -2, "name" );

    EXPECT_FLOAT_EQ( 2000, get_lua_linear( L, "pfx", "name", 0 ) );


}

TEST_F( PhysQuant, get_lua_extent_bad_unit ) {

    lua_atpanic( L, panic_noerr );

    lua_newtable( L );  // container

    lua_newtable( L );	// physquant table
    int table = lua_gettop( L );

    lua_pushinteger( L, 1 );	// key
    lua_pushnumber( L, 2 );	// value
    lua_settable( L, table );

    ut_unit* keV = raygen::Units::parse( "keV" );

    lua_pushinteger( L, 2 );	// key
    luaut_push_udata_unit(L, keV, 0);  // value
    lua_settable( L, table );

    lua_setfield( L, -2, "frank" );

    try {
	get_lua_extent( L, "test", "frank", 2000, raygen::Units::sys().mm());
	FAIL();
    }
    catch( std::string &err ) {

	ASSERT_NE( std::string::npos, err.find( " must be a linear or angular" ) );
    }
    catch( ... ) {

	FAIL();

    }
}


TEST_F( PhysQuant, get_lua_extent_no_unit ) {

    lua_atpanic( L, panic_noerr );

    lua_newtable( L );  // container

    lua_pushnumber( L, 2 );	// value

    lua_setfield( L, -2, "frank" );

    try {
	get_lua_extent( L, "test", "frank", 2000, raygen::Units::sys().mm());
	FAIL();
    }
    catch( std::string &err ) {

	ASSERT_NE( std::string::npos, err.find( "unit must be specified" ) );
    }
    catch( ... ) {

	FAIL();

    }

}


TEST_F( PhysQuant, get_lua_extent_angular ) {

    lua_newtable( L );  // container

    lua_newtable( L );	// physquant table
    int table = lua_gettop( L );

    lua_pushinteger( L, 1 );	// key
    lua_pushnumber( L, 2 );	// value
    lua_settable( L, table );

    ut_unit* meter = raygen::Units::parse( "meter" );

    lua_pushinteger( L, 2 );	// key
    luaut_push_udata_unit(L, meter, 0);  // value
    lua_settable( L, table );

    lua_setfield( L, -2, "frank" );

    double extent;

    extent = get_lua_extent( L, "test", "frank", 2000,
			     raygen::Units::sys().mm() );
    EXPECT_FLOAT_EQ( 2000, extent );

    ut_unit* degree = raygen::Units::parse( "degree" );
    extent = get_lua_extent( L, "test", "frank", 2000, degree );
    EXPECT_FLOAT_EQ( 45, extent );

    ut_free( meter );
    ut_free( degree );

}

TEST_F( PhysQuant, get_lua_extent_linear ) {

    lua_newtable( L );  // container

    lua_newtable( L );	// physquant table
    int table = lua_gettop( L );

    lua_pushinteger( L, 1 );	// key
    lua_pushnumber( L, 45 );	// value
    lua_settable( L, table );

    ut_unit* degree = raygen::Units::parse( "degree" );

    lua_pushinteger( L, 2 );	// key
    luaut_push_udata_unit(L, degree, 0);  // value
    lua_settable( L, table );

    lua_setfield( L, -2, "frank" );

    double extent;

    extent = get_lua_extent( L, "test", "frank", 2000,
			     degree );
    EXPECT_FLOAT_EQ( 45, extent );

    extent = get_lua_extent( L, "test", "frank", 2000,
			     raygen::Units::sys().mm() );
    EXPECT_FLOAT_EQ( 2000, extent );

}

